"""Fabric Warehouse job client implementation - based on Synapse with COPY INTO support"""

import os
from typing import ClassVar, Sequence, List, Dict
from copy import deepcopy
from textwrap import dedent
from urllib.parse import urlparse

from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.schema.typing import TColumnSchema
from dlt.common.schema import Schema
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import LoadJob
from dlt.destinations.impl.synapse.synapse import (
    SynapseClient,
    HINT_TO_SYNAPSE_ATTR,
    SynapseCopyFileLoadJob,
)
from dlt.destinations.impl.fabric.configuration import FabricClientConfiguration
from dlt.destinations.impl.fabric.sql_client import FabricSqlClient
from dlt.destinations.job_client_impl import SqlJobClientBase
from dlt.common.configuration.exceptions import ConfigurationException
from dlt.common.configuration.specs import (
    AzureCredentialsWithoutDefaults,
    AzureServicePrincipalCredentialsWithoutDefaults,
)


class FabricCopyFileLoadJob(SynapseCopyFileLoadJob):
    """Custom COPY INTO job for Fabric that removes AUTO_CREATE_TABLE parameter"""

    # Class-level cache for initialized Service Principal tokens to avoid rate limiting
    _token_initialized_cache: ClassVar[Dict[str, bool]] = {}

    def run(self) -> None:
        self._sql_client = self._job_client.sql_client
        # Get file type
        ext = os.path.splitext(self._bucket_path)[1][1:]
        if ext == "parquet":
            file_type = "PARQUET"
        else:
            raise ValueError(f"Unsupported file type `{ext}` for Fabric.")

        staging_credentials = self._staging_credentials
        assert staging_credentials is not None
        assert isinstance(
            staging_credentials,
            (AzureCredentialsWithoutDefaults, AzureServicePrincipalCredentialsWithoutDefaults),
        )
        azure_storage_account_name = staging_credentials.azure_storage_account_name
        https_path = self._get_https_path(self._bucket_path, azure_storage_account_name)
        table_name = self._load_table["name"]

        # Check if this is OneLake storage
        bucket_url = urlparse(self._bucket_path)
        is_onelake = (
            bucket_url.scheme == "abfss" and "onelake.dfs.fabric.microsoft.com" in bucket_url.netloc
        )

        # For OneLake with Service Principal, initialize Fabric token first
        # Token initialization is cached to avoid rate limiting on multiple file loads
        if is_onelake and isinstance(
            staging_credentials, AzureServicePrincipalCredentialsWithoutDefaults
        ):
            self._ensure_fabric_token_initialized(staging_credentials)

        # Build WITH clause options
        with_options = [f"FILE_TYPE = '{file_type}'"]

        if not is_onelake:
            # For Azure Storage (non-OneLake), add credential
            if isinstance(staging_credentials, AzureCredentialsWithoutDefaults):
                sas_token = staging_credentials.azure_storage_sas_token
                credential = f"IDENTITY = 'Shared Access Signature', SECRET = '{sas_token}'"
            else:
                raise ConfigurationException(
                    f"Credentials of type `{type(staging_credentials)}` not supported"
                    " when loading data from staging into Fabric using `COPY INTO`."
                )
            with_options.append(f"CREDENTIAL = ({credential})")

        # Copy data from staging file into Fabric table
        with self._sql_client.begin_transaction():
            dataset_name = self._sql_client.dataset_name
            with_clause = ",\n                    ".join(with_options)
            sql = dedent(f"""
                COPY INTO [{dataset_name}].[{table_name}]
                FROM '{https_path}'
                WITH (
                    {with_clause}
                )
            """)
            self._sql_client.execute_sql(sql)

    def _ensure_fabric_token_initialized(
        self, credentials: AzureServicePrincipalCredentialsWithoutDefaults
    ) -> None:
        """
        Ensure Fabric token is initialized for Service Principal, using cache to avoid rate limiting.
        This is required before the SP can access OneLake through COPY INTO or OPENROWSET.

        Token initialization is cached per client_id to prevent excessive API calls during bulk loads.
        """
        cache_key = credentials.azure_client_id

        # Check if we've already initialized the token for this client
        if cache_key in self._token_initialized_cache:
            return

        try:
            import requests
            from azure.identity import ClientSecretCredential
        except ImportError:
            raise ConfigurationException(
                "azure-identity and requests packages are required for Service Principal "
                "authentication with Fabric OneLake. Install them with: "
                "pip install azure-identity requests"
            )

        # Create credential and get token
        cred = ClientSecretCredential(
            tenant_id=credentials.azure_tenant_id,
            client_id=credentials.azure_client_id,
            client_secret=credentials.azure_client_secret,
        )
        token = cred.get_token("https://api.fabric.microsoft.com/.default").token

        # Call Fabric API to initialize token (list workspaces as a simple test)
        headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}
        resp = requests.get("https://api.fabric.microsoft.com/v1/workspaces", headers=headers)

        if resp.status_code != 200:
            raise ConfigurationException(
                "Failed to initialize Fabric token for Service Principal. "
                f"Status: {resp.status_code}, Response: {resp.text}"
            )

        # Cache successful initialization
        self._token_initialized_cache[cache_key] = True

    def _get_https_path(self, bucket_path: str, storage_account_name: str) -> str:
        """
        For OneLake paths (abfss://), convert to https://onelake.dfs.fabric.microsoft.com format.
        For regular Azure Storage (az://), use the standard blob endpoint conversion.
        """
        bucket_url = urlparse(bucket_path)

        # Check if this is a OneLake path (abfss:// scheme or onelake in the hostname)
        if bucket_url.scheme == "abfss" and "onelake.dfs.fabric.microsoft.com" in bucket_url.netloc:
            # OneLake abfss path: abfss://workspace@onelake.dfs.fabric.microsoft.com/item/path/file
            # Convert to: https://onelake.dfs.fabric.microsoft.com/workspace/item/path/file
            workspace = bucket_url.netloc.split("@")[0]
            path = bucket_url.path
            return f"https://onelake.dfs.fabric.microsoft.com/{workspace}{path}"
        else:
            # Regular Azure Storage path - use parent implementation
            return super()._get_https_path(bucket_path, storage_account_name)


class FabricClient(SynapseClient):
    """Custom job client for Fabric Warehouse that uses varchar instead of nvarchar and supports COPY INTO"""

    def __init__(
        self,
        schema: Schema,
        config: FabricClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        # Call grandparent init (MsSqlJobClient) but set up Fabric-specific client
        # We'll initialize our own sql_client below
        super(SynapseClient, self).__init__(schema, config, capabilities)  # type: ignore[arg-type]
        self.config: FabricClientConfiguration = config  # type: ignore[assignment]

        # Create Fabric-specific SQL client
        from dlt.destinations.impl.mssql.mssql import MsSqlJobClient

        dataset_name, staging_dataset_name = MsSqlJobClient.create_dataset_names(schema, config)
        self.sql_client = FabricSqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
        )

        self.active_hints = deepcopy(HINT_TO_SYNAPSE_ATTR)
        if not self.config.create_indexes:
            self.active_hints.pop("primary_key", None)
            self.active_hints.pop("unique", None)

    def _get_column_def_sql(self, c: TColumnSchema, table: PreparedTableSchema = None) -> str:
        """Override to use varchar instead of nvarchar for unique text columns"""
        sc_type = c["data_type"]
        if sc_type == "text" and c.get("unique"):
            # Fabric does not support nvarchar - use varchar with max length 900 for unique columns
            db_type = "varchar(%i)" % (c.get("precision") or 900)
        else:
            db_type = self.type_mapper.to_destination_type(c, table)

        # Don't add COLLATE clause here - let the database default handle it
        # The warehouse-level collation will be applied automatically

        hints_str = self._get_column_hints_sql(c)
        column_name = self.sql_client.escape_column_name(c["name"])
        return f"{column_name} {db_type} {hints_str} {self._gen_not_null(c.get('nullable', True))}"

    def prepare_load_table(self, table_name: str) -> PreparedTableSchema:
        """Override to ensure proper table configuration for Fabric

        Note: Fabric doesn't support table indexing - it automatically manages storage.
        """
        table = super(SynapseClient, self).prepare_load_table(table_name)

        # For _dlt_version table, convert all text columns to varchar(max) to avoid
        # pyodbc binding them as legacy text/ntext types which don't support UTF-8 collations
        if table_name == self.schema.version_table_name:
            for column in table["columns"].values():
                if column.get("data_type") == "text":
                    # Override type mapper behavior for this specific table
                    # Use varchar(max) with explicit precision to avoid text/ntext binding
                    column["precision"] = 2147483647  # max value for varchar(max)

        return table

    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        return self.config.truncate_tables_on_staging_destination_before_load

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        """Override to handle file loading - Fabric requires staging for parquet files

        Fabric doesn't use ADBC for direct parquet loading. Instead, it requires staging
        storage (OneLake or Azure Blob) and uses COPY INTO for efficient bulk loading.
        """
        from dlt.common.storages.load_package import ParsedLoadJobFileName
        from dlt.destinations.job_impl import ReferenceFollowupJobRequest

        # Check for reference jobs (staging files)
        if ReferenceFollowupJobRequest.is_reference_job(file_path):
            job = FabricCopyFileLoadJob(
                file_path,
                self.config.staging_config.credentials,  # type: ignore[arg-type]
            )
            return job

        # For non-reference jobs, only handle insert-values and sql files
        parsed_file = ParsedLoadJobFileName.parse(file_path)
        if parsed_file.file_format in ("insert_values", "sql", "model"):
            # Use parent's insert job handling
            return super(SynapseClient, self).create_load_job(table, file_path, load_id, restore)

        # Parquet and other file formats require staging
        raise ValueError(
            f"Fabric requires staging storage for {parsed_file.file_format} files. "
            "Configure staging with filesystem destination (OneLake or Azure Blob Storage)."
        )

    def _get_table_update_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> List[str]:
        """Override to remove WITH clause - Fabric automatically manages storage"""
        # Fabric doesn't support WITH (HEAP) or WITH (CLUSTERED COLUMNSTORE INDEX)
        # Just return the base SQL without any WITH clause
        return SqlJobClientBase._get_table_update_sql(self, table_name, new_columns, generate_alter)



from contextlib import contextmanager, suppress
import os
import tempfile
import warnings
from types import ModuleType
from typing import Any, Dict, Iterator, List, Optional
from urllib.parse import urlencode

from packaging.specifiers import SpecifierSet

from dlt.common import known_env
from dlt.common.configuration.container import Container
from dlt.common.configuration.providers import (
    EnvironProvider,
    SecretsTomlProvider,
    ConfigTomlProvider,
)
from dlt.common.configuration.providers.provider import ConfigProvider
from dlt.common.configuration.resolve import resolve_configuration
from dlt.common.configuration.specs.base_configuration import BaseConfiguration
from dlt.common.configuration.specs.pluggable_run_context import (
    RunContextBase,
    PluggableRunContext,
)
from dlt.common.configuration.specs.runtime_configuration import RuntimeConfiguration
from dlt.common.runtime.init import initialize_runtime

# dlt settings folder
DOT_DLT = os.environ.get(known_env.DLT_CONFIG_FOLDER, ".dlt")


class RunContext(RunContextBase):
    """A default run context used by dlt"""

    def __init__(self, run_dir: Optional[str]):
        self._init_run_dir = run_dir or "."
        self._runtime_config: RuntimeConfiguration = None

    @property
    def global_dir(self) -> str:
        return global_dir()

    @property
    def uri(self) -> str:
        return context_uri(self.name, self.run_dir, self.runtime_kwargs)

    @property
    def run_dir(self) -> str:
        """The default run dir is the current working directory but may be overridden by DLT_PROJECT_DIR env variable."""
        return os.environ.get(known_env.DLT_PROJECT_DIR, self._init_run_dir)

    @property
    def local_dir(self) -> str:
        return os.environ.get(known_env.DLT_LOCAL_DIR, self._init_run_dir)

    @property
    def settings_dir(self) -> str:
        """Returns a path to dlt settings directory. If not overridden it resides in current working directory

        The name of the setting folder is '.dlt'. The path is current working directory '.' but may be overridden by DLT_PROJECT_DIR env variable.
        """
        return os.path.join(self.run_dir, DOT_DLT)

    @property
    def data_dir(self) -> str:
        return os.environ.get(known_env.DLT_DATA_DIR, global_dir())

    def initial_providers(self) -> List[ConfigProvider]:
        providers = [
            EnvironProvider(),
            SecretsTomlProvider(self.settings_dir, self.global_dir),
            ConfigTomlProvider(self.settings_dir, self.global_dir),
        ]
        return providers

    def initialize_runtime(self, runtime_config: RuntimeConfiguration = None) -> None:
        if runtime_config is None:
            self._runtime_config = resolve_configuration(RuntimeConfiguration())
        else:
            self._runtime_config = runtime_config

        initialize_runtime(self.name, self._runtime_config)

    @property
    def runtime_config(self) -> RuntimeConfiguration:
        if self._runtime_config is None:
            self.initialize_runtime()
        return self._runtime_config

    @property
    def config(self) -> BaseConfiguration:
        return None

    @property
    def module(self) -> Optional[ModuleType]:
        try:
            return self.import_run_dir_module(self.run_dir)
        except (ImportError, TypeError):
            return None

    @property
    def runtime_kwargs(self) -> Dict[str, Any]:
        return None

    def get_data_entity(self, entity: str) -> str:
        return os.path.join(self.data_dir, entity)

    def get_run_entity(self, entity: str) -> str:
        """Default run context assumes that entities are defined in root dir"""
        return self.run_dir

    def get_setting(self, setting_path: str) -> str:
        return os.path.join(self.settings_dir, setting_path)

    def plug(self) -> None:
        pass

    def unplug(self) -> None:
        pass

    def reset_config(self) -> None:
        self._runtime_config = None

    @property
    def name(self) -> str:
        return "dlt"


def switch_context(
    run_dir: Optional[str], profile: str = None, required: str = None, validate: bool = False
) -> RunContextBase:
    """Switch the run context to a project at `run_dir` with an optional profile.

    Calls `reload` on `PluggableRunContext` to re-trigger the plugin hook
    (`plug_run_context` spec), which will query all active context plugins.

    The `required` argument is passed to each context plugin via the
    `_required` key of `runtime_kwargs` and should cause an exception if a
    given plugin cannot instantiate its context at `run_dir`.

    The `validate` argument is passed to each context plugin via the
    `_validate` key of `runtime_kwargs` and should cause a strict validation
    of any config files and manifests associated with the run context.

    Args:
        run_dir (str): Filesystem path of the project directory to activate. If None,
            plugins may resolve the directory themselves.
        profile (str): Profile name to activate for the run context.
        required (str, optional): A class name of the context be instantiated at `run_dir` ie. setting
            it to `WorkspaceRunContext` will cause the workspace context plugin to raise if workspace is
            not found at `run_dir`.
        validate (str, optional): If True, plugins should perform strict validation of config
            files and manifests associated with the run context.

    Returns:
        SupportsProfilesRunContext: The new run context.
    """
    container = Container()
    # reload run context via plugins
    container[PluggableRunContext].reload(
        run_dir, dict(profile=profile, _required=required, _validate=validate)
    )
    # return new run context
    return container[PluggableRunContext].context


@contextmanager
def switched_run_context(new_context: RunContext) -> Iterator[RunContext]:
    """Context manager that switches run context to `new_context` into pluggable run context."""
    container = Container()
    cookie = container[PluggableRunContext].push_context()
    try:
        container[PluggableRunContext].reload(new_context)
        yield new_context
    finally:
        container[PluggableRunContext].pop_context(cookie)


def global_dir() -> str:
    """Gets default directory where pipelines' data (working directories) will be stored
    1. if XDG_DATA_HOME is set in env then it is used
    2. in user home directory: ~/.dlt/
    3. if current user is root: in /var/dlt/
    4. if current user does not have a home directory: in /tmp/dlt/
    """
    # geteuid not available on Windows
    if hasattr(os, "geteuid") and os.geteuid() == 0:
        # we are root so use standard /var
        return os.path.join("/var", "dlt")

    home = os.path.expanduser("~")
    if home is None or not is_folder_writable(home):
        # no home dir - use temp
        return os.path.join(tempfile.gettempdir(), "dlt")
    else:
        # use XDG_DATA_HOME only if ~/.dlt doesn't exist
        if "XDG_DATA_HOME" in os.environ:
            if not os.path.isdir(os.path.join(home, DOT_DLT)):
                return os.path.join(os.environ["XDG_DATA_HOME"], "dlt")
            else:
                warnings.warn(
                    f"XDG_DATA_HOME is set to {os.environ['XDG_DATA_HOME']} but ~/.dlt already"
                    " exists. Using ~/.dlt"
                )

        # else default to ~/.dlt
        return os.path.join(home, DOT_DLT)


def is_folder_writable(path: str) -> bool:
    import tempfile

    try:
        # Ensure the path exists
        if not os.path.exists(path):
            return False
        # Attempt to create a temporary file
        with tempfile.TemporaryFile(dir=path):
            pass
        return True
    except OSError:
        return False


def get_plugin_modules() -> List[str]:
    """Return top level module names of all discovered plugins, including `dlt`.

    If current run context is a top level module it is also included, otherwise empty string.
    """
    from dlt.common.configuration.plugins import PluginContext

    # get current run module
    ctx_module = active().module
    run_module_name = ctx_module.__name__ if ctx_module else ""

    plugin_modules = Container()[PluginContext].plugin_modules.copy()
    with suppress(ValueError):
        plugin_modules.remove(run_module_name)
    plugin_modules.insert(0, run_module_name)
    return plugin_modules


def ensure_plugin_version_match(
    pkg_name: str,
    dlt_version: str,
    plugin_version: str,
    plugin_module_name: str,
    dlt_extra: str,
    dlt_version_specifier: Optional[SpecifierSet] = None,
) -> None:
    """Ensures that installed plugin version matches dlt requirements. Plugins are tightly bound
    to `dlt` and released together.

    If `dlt_version_specifier` is provided, it is used to check if the plugin version satisfies
    the specifier. Otherwise, the specifier is read from dlt's package metadata (Requires-Dist).
    If specifier cannot be determined, the function returns without checking.

    Args:
        pkg_name: Name of the plugin package (e.g., "dlthub")
        dlt_version: The installed dlt version string
        plugin_version: The installed plugin version string
        plugin_module_name: The module name for MissingDependencyException (e.g., "dlthub")
        dlt_extra: The dlt extra to install the plugin (e.g., "hub")
        dlt_version_specifier: Optional version specifier for the plugin. If not provided,
            reads from dlt's package metadata.

    Raises:
        MissingDependencyException: If version mismatch is detected
    """
    # Get specifier from dlt's package metadata if not provided
    if dlt_version_specifier is None:
        from dlt.version import get_dependency_requirement

        req = get_dependency_requirement(pkg_name)
        if req is not None:
            dlt_version_specifier = req.specifier

    # If specifier still not available, exit without checking
    if dlt_version_specifier is None or len(dlt_version_specifier) == 0:
        return

    # Use specifier.contains() for proper version check (allowing prereleases)
    if not dlt_version_specifier.contains(plugin_version, prereleases=True):
        from dlt.common.exceptions import MissingDependencyException

        custom_msg = (
            f"`{pkg_name}` is a `dlt` plugin and must satisfy version requirement "
            f"`{dlt_version_specifier}` but you have {plugin_version}. "
            f"Please install the right version of {pkg_name} with:\n\n"
            f'pip install "dlt[{dlt_extra}]=={dlt_version}"\n\n'
            "or if you are upgrading the plugin:\n\n"
            f'pip install "dlt[{dlt_extra}]=={dlt_version}" -U {pkg_name}'
        )
        missing_dep_ex = MissingDependencyException(plugin_module_name, [])
        missing_dep_ex.args = (custom_msg,)
        missing_dep_ex.msg = custom_msg
        raise missing_dep_ex


def context_uri(name: str, run_dir: str, runtime_kwargs: Optional[Dict[str, Any]]) -> str:
    from dlt.common.storages.configuration import FilesystemConfiguration

    uri_no_qs = FilesystemConfiguration.make_file_url(run_dir)
    # add query string from self.runtime_kwargs
    if runtime_kwargs:
        # skip kwargs starting with _
        query_string = urlencode({k: v for k, v in runtime_kwargs.items() if not k.startswith("_")})
        if query_string:
            return f"{uri_no_qs}?{query_string}"
    return uri_no_qs


def active() -> RunContextBase:
    """Returns currently active run context"""
    return Container()[PluggableRunContext].context



from .pool_runner import run_pool, NullExecutor
from .runnable import Runnable, workermethod, TExecutor
from .typing import TRunMetrics
from .venv import Venv, VenvNotFound


__all__ = [
    "run_pool",
    "NullExecutor",
    "Runnable",
    "workermethod",
    "TExecutor",
    "TRunMetrics",
    "Venv",
    "VenvNotFound",
]



import os
from abc import abstractmethod
import base64
import contextlib
from copy import copy
from types import TracebackType
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Iterable,
    Iterator,
)
import zlib
import re

import sqlglot.expressions

from dlt.common.libs.sqlglot import TSqlGlotDialect
from dlt.common import pendulum, logger
from dlt.common.destination.capabilities import DataTypeMapper
from dlt.common.destination.exceptions import WriteDispositionNotSupported
from dlt.common.destination.utils import resolve_replace_strategy
from dlt.common.json import json
from dlt.common.schema.typing import (
    C_DLT_LOAD_ID,
    C_DLT_LOADS_TABLE_LOAD_ID,
    COLUMN_HINTS,
    TColumnType,
    TColumnSchemaBase,
    TPartialTableSchema,
)
from dlt.common.schema.utils import (
    get_inherited_table_hint,
    has_default_column_prop_value,
    loads_table,
    normalize_table_identifiers,
    version_table,
)
from dlt.common.utils import read_dialect_and_sql
from dlt.common.storages import FileStorage
from dlt.common.storages.load_package import LoadJobInfo, ParsedLoadJobFileName
from dlt.common.schema import TColumnSchema, Schema, TTableSchemaColumns, TSchemaTables
from dlt.common.schema import TColumnHint
from dlt.common.destination.client import (
    PreparedTableSchema,
    StateInfo,
    StorageSchemaInfo,
    WithStateSync,
    DestinationClientConfiguration,
    DestinationClientDwhConfiguration,
    FollowupJobRequest,
    WithStagingDataset,
    RunnableLoadJob,
    LoadJob,
    JobClientBase,
    HasFollowupJobs,
    CredentialsConfiguration,
)

from dlt.destinations.exceptions import DatabaseUndefinedRelation
from dlt.destinations.job_impl import (
    ReferenceFollowupJobRequest,
)
from dlt.destinations.sql_jobs import SqlMergeFollowupJob, SqlStagingReplaceFollowupJob
from dlt.destinations.typing import TNativeConn
from dlt.destinations.sql_client import SqlClientBase, WithSqlClient
from dlt.destinations.utils import (
    get_pipeline_state_query_columns,
    info_schema_null_to_bool,
    verify_schema_merge_disposition,
    verify_schema_replace_disposition,
)

import sqlglot

# this should suffice for now
DDL_COMMANDS = ["ALTER", "CREATE", "DROP"]
# TODO: move to respective sql clients
UNLOGGED_COMMANDS = ["ALTER SCHEMA"]


class SqlLoadJob(RunnableLoadJob):
    """A job executing sql statement, without followup trait"""

    def __init__(self, file_path: str) -> None:
        super().__init__(file_path)
        self._job_client: "SqlJobClientBase" = None

    def run(self) -> None:
        self._sql_client = self._job_client.sql_client
        # execute immediately if client present
        with FileStorage.open_zipsafe_ro(self._file_path, "r", encoding="utf-8") as f:
            sql = f.read()

        with self.maybe_transaction(sql):
            # Some clients (e.g. databricks) do not support multiple statements in one execute call
            self._sql_client.execute_many(self._split_fragments(sql))

    @contextlib.contextmanager
    def maybe_transaction(self, sql: str) -> Iterator[None]:
        """Begins a transaction if sql client supports it, otherwise works in auto commit."""
        is_ddl_statement = self._string_contains_ddl_statements(sql)
        if (
            self._job_client.capabilities.supports_ddl_transactions
            and is_ddl_statement
            or self._job_client.capabilities.supports_transactions
            and not is_ddl_statement
        ) and not self._has_out_of_transaction_commands(sql):
            with self._sql_client.begin_transaction():
                yield
        else:
            yield

    def _string_contains_ddl_statements(self, sql: str) -> bool:
        for cmd in DDL_COMMANDS:
            if re.search(cmd, sql, re.IGNORECASE):
                return True
        return False

    def _has_out_of_transaction_commands(self, sql: str) -> bool:
        for cmd in UNLOGGED_COMMANDS:
            if re.search(cmd, sql, re.IGNORECASE):
                return True
        # disable transaction on synapse when doing sql jobs, to enable:
        # TODO: 1. disable transaction if temp table is created via select into
        # 2. swap TRUNCATE for delete in replace jobs
        if "SynapseClient" in self._job_client.__class__.__name__:
            return True
        return False

    def _split_fragments(self, sql: str) -> List[str]:
        return [s.strip() for line in sql.split("\n") for s in line.split(";") if s.strip()]

    @staticmethod
    def is_sql_job(file_path: str) -> bool:
        return os.path.splitext(file_path)[1][1:] == "sql"


class ModelLoadJob(RunnableLoadJob, HasFollowupJobs):
    """
    A job to insert rows into a table from a model file which contains a single select statement
    """

    def __init__(self, file_path: str) -> None:
        super().__init__(file_path)
        self._job_client: "SqlJobClientBase" = None

    def run(self) -> None:
        with FileStorage.open_zipsafe_ro(self._file_path, "r", encoding="utf-8") as f:
            select_dialect, select_statement = read_dialect_and_sql(
                file_obj=f,
                fallback_dialect=self._job_client.capabilities.sqlglot_dialect,  # caps are available at this point
            )

        sql_client = self._job_client.sql_client
        insert_statement = self._insert_statement_from_select_statement(
            select_dialect, select_statement
        )
        sql_client.execute_sql(insert_statement)

    def _insert_statement_from_select_statement(
        self, select_dialect: TSqlGlotDialect, select_statement: str
    ) -> str:
        """
        Generates an INSERT statement from a SELECT statement using sqlglot.
        Transpiles the SELECT if needed and adjusts identifiers for the destination dialect.
        """
        sql_client = self._job_client.sql_client
        target_table = sql_client.make_qualified_table_name(self._load_table["name"])
        target_catalog = sql_client.catalog_name(quote=False)
        destination_dialect = self._job_client.capabilities.sqlglot_dialect

        # Parse SELECT
        parsed_select = sqlglot.parse_one(select_statement, read=select_dialect)

        # Adjust table parts (catalog/db/this) based on dialect and catalog presence
        if select_dialect != destination_dialect:
            # TODO: We might need this
            for table in parsed_select.find_all(sqlglot.exp.Table):
                parts = list(table.parts)
                if target_catalog:
                    if len(parts) == 3:
                        table.set("catalog", sqlglot.to_identifier(target_catalog))
                        table.set("db", sqlglot.to_identifier(parts[1].name))
                        table.set("this", sqlglot.to_identifier(parts[2].name))
                    elif len(parts) == 2:
                        table.set("catalog", sqlglot.to_identifier(target_catalog))
                        table.set("db", sqlglot.to_identifier(parts[0].name))
                        table.set("this", sqlglot.to_identifier(parts[1].name))
                else:
                    if len(parts) == 3:
                        table.set("catalog", None)
                        table.set("db", sqlglot.to_identifier(parts[1].name))
                        table.set("this", sqlglot.to_identifier(parts[2].name))

        # Ensure there's a top-level SELECT, otherwise it doesn't make sense
        top_level_select = parsed_select.find(sqlglot.exp.Select)
        if top_level_select is None:
            raise ValueError(
                "No top-level SELECT statement found in the model query. Models must be defined"
                " using a SELECT statement."
            )

        # Get identifiers for the columns in the INSERT statement
        # Normalizer aliased all selections
        columns = []
        for _, expr in enumerate(top_level_select.expressions):
            alias_name = expr.alias
            columns.append(sqlglot.to_identifier(alias_name, quoted=True))

        # Build final INSERT
        query = sqlglot.expressions.insert(
            expression=parsed_select,
            into=target_table,
            columns=columns,
            dialect=destination_dialect,
        ).sql(destination_dialect)

        return query

    @staticmethod
    def is_model_job(file_path: str) -> bool:
        return os.path.splitext(file_path)[1][1:] == "model"


class CopyRemoteFileLoadJob(RunnableLoadJob, HasFollowupJobs):
    def __init__(
        self,
        file_path: str,
        staging_credentials: Optional[CredentialsConfiguration] = None,
    ) -> None:
        super().__init__(file_path)
        self._job_client: "SqlJobClientBase" = None
        self._staging_credentials = staging_credentials
        self._bucket_path = ReferenceFollowupJobRequest.resolve_reference(file_path)


class SqlJobClientBase(WithSqlClient, JobClientBase, WithStateSync):
    def __init__(
        self,
        schema: Schema,
        config: DestinationClientConfiguration,
        sql_client: SqlClientBase[TNativeConn],
    ) -> None:
        # get definitions of the dlt tables, normalize column names and keep for later use
        version_table_ = normalize_table_identifiers(version_table(), schema.naming)
        self.version_table_schema_columns = ", ".join(
            sql_client.escape_column_name(col) for col in version_table_["columns"]
        )
        loads_table_ = normalize_table_identifiers(loads_table(), schema.naming)
        self.loads_table_schema_columns = ", ".join(
            sql_client.escape_column_name(col) for col in loads_table_["columns"]
        )
        state_table_ = normalize_table_identifiers(
            get_pipeline_state_query_columns(), schema.naming
        )
        self.state_table_columns = ", ".join(
            sql_client.escape_column_name(col) for col in state_table_["columns"]
        )
        self.active_hints: Dict[TColumnHint, str] = {}
        self.type_mapper: DataTypeMapper = None
        self.allow_merge_to_append_fallback: bool = False
        super().__init__(schema, config, sql_client.capabilities)
        self.sql_client = sql_client
        assert isinstance(config, DestinationClientDwhConfiguration)
        self.config: DestinationClientDwhConfiguration = config

    @property
    def sql_client_class(self) -> Type[SqlClientBase[Any]]:
        return self._sql_client.__class__

    @property
    def sql_client(self) -> SqlClientBase[TNativeConn]:
        return self._sql_client

    @sql_client.setter
    def sql_client(self, client: SqlClientBase[TNativeConn]) -> None:
        self._sql_client = client

    def drop_storage(self) -> None:
        self.sql_client.drop_dataset()
        with contextlib.suppress(DatabaseUndefinedRelation):
            with self.sql_client.with_staging_dataset():
                self.sql_client.drop_dataset()

    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        if not self.is_storage_initialized():
            self.sql_client.create_dataset()
        elif truncate_tables:
            self.sql_client.truncate_tables(*truncate_tables)

    def is_storage_initialized(self) -> bool:
        return self.sql_client.has_dataset()

    def update_stored_schema(
        self,
        only_tables: Iterable[str] = None,
        expected_update: TSchemaTables = None,
    ) -> Optional[TSchemaTables]:
        super().update_stored_schema(only_tables, expected_update)
        applied_update: TSchemaTables = {}
        schema_info = self.get_stored_schema_by_hash(self.schema.stored_version_hash)
        if schema_info is None:
            logger.info(
                f"Schema with hash {self.schema.stored_version_hash} not found in the storage."
                " upgrading"
            )

            with self.maybe_ddl_transaction():
                applied_update = self._execute_schema_update_sql(only_tables)
        else:
            logger.info(
                f"Schema with hash {self.schema.stored_version_hash} inserted at"
                f" {schema_info.inserted_at} found in storage, no upgrade required"
            )
        return applied_update

    def drop_tables(self, *tables: str, delete_schema: bool = True) -> None:
        """Drop tables in destination database and optionally delete the stored schema as well.
        Clients that support ddl transactions will have both operations performed in a single transaction.

        Args:
            tables: Names of tables to drop.
            delete_schema: If True, also delete all versions of the current schema from storage
        """
        with self.maybe_ddl_transaction():
            self.sql_client.drop_tables(*tables)
            if delete_schema:
                self._delete_schema_in_storage(self.schema)

    @contextlib.contextmanager
    def maybe_ddl_transaction(self) -> Iterator[None]:
        """Begins a transaction if sql client supports it, otherwise works in auto commit."""
        if self.capabilities.supports_ddl_transactions:
            with self.sql_client.begin_transaction():
                yield
        else:
            yield

    def prepare_load_table(self, table_name: str) -> PreparedTableSchema:
        load_table = super().prepare_load_table(table_name)
        # apply x-replace-strategy (still unofficial)
        if load_table["write_disposition"] == "replace":
            replace_strategy = resolve_replace_strategy(
                load_table, self.config.replace_strategy, self.capabilities
            )
            assert replace_strategy, f"Must be able to get replace strategy for {table_name}"
            load_table["x-replace-strategy"] = replace_strategy  # type: ignore[typeddict-unknown-key]
        return load_table

    def should_truncate_table_before_load(self, table_name: str) -> bool:
        table = self.prepare_load_table(table_name)
        return (
            table["write_disposition"] == "replace"
            and table["x-replace-strategy"] == "truncate-and-insert"  # type: ignore[typeddict-item]
        )

    def _create_append_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return []

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return [SqlMergeFollowupJob.from_table_chain(table_chain, self.sql_client)]

    def _create_replace_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        jobs: List[FollowupJobRequest] = []
        root_table = table_chain[0]
        if root_table["x-replace-strategy"] in ["insert-from-staging", "staging-optimized"]:  # type: ignore[typeddict-item]
            jobs.append(SqlStagingReplaceFollowupJob.from_table_chain(table_chain, self.sql_client))
        return jobs

    def create_table_chain_completed_followup_jobs(
        self,
        table_chain: Sequence[PreparedTableSchema],
        completed_table_chain_jobs: Optional[Sequence[LoadJobInfo]] = None,
    ) -> List[FollowupJobRequest]:
        """Creates a list of followup jobs for merge write disposition and staging replace strategies"""
        jobs = super().create_table_chain_completed_followup_jobs(
            table_chain, completed_table_chain_jobs
        )
        write_disposition = table_chain[0]["write_disposition"]
        if write_disposition == "append":
            jobs.extend(self._create_append_followup_jobs(table_chain))
        elif write_disposition == "merge":
            jobs.extend(self._create_merge_followup_jobs(table_chain))
        elif write_disposition == "replace":
            jobs.extend(self._create_replace_followup_jobs(table_chain))
        return jobs

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        """Starts SqlLoadJob for files ending with .sql or returns None to let derived classes to handle their specific jobs"""
        if SqlLoadJob.is_sql_job(file_path):
            # create sql load job
            return SqlLoadJob(file_path)
        if ModelLoadJob.is_model_job(file_path):
            # create model load job
            return ModelLoadJob(file_path)
        return None

    def complete_load(self, load_id: str) -> None:
        name = self.sql_client.make_qualified_table_name(self.schema.loads_table_name)
        now_ts = pendulum.now()
        self.sql_client.execute_sql(
            f"INSERT INTO {name}({self.loads_table_schema_columns}) VALUES(%s, %s, %s, %s, %s)",
            load_id,
            self.schema.name,
            0,
            now_ts,
            self.schema.version_hash,
        )

    def __enter__(self) -> "SqlJobClientBase":
        self.sql_client.open_connection()
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        self.sql_client.close_connection()

    def get_storage_tables(
        self, table_names: Iterable[str]
    ) -> Iterable[Tuple[str, TTableSchemaColumns]]:
        """Uses INFORMATION_SCHEMA to retrieve table and column information for tables in `table_names` iterator.
        Table names should be normalized according to naming convention and will be further converted to desired casing
        in order to (in most cases) create case-insensitive name suitable for search in information schema.

        The column names are returned as in information schema. To match those with columns in existing table, you'll need to use
        `schema.get_new_table_columns` method and pass the correct casing. Most of the casing function are irreversible so it is not
        possible to convert identifiers into INFORMATION SCHEMA back into case sensitive dlt schema.
        """
        table_names = list(table_names)
        if len(table_names) == 0:
            # empty generator
            return
        # get schema search components
        catalog_name, schema_name, folded_table_names = (
            self.sql_client._get_information_schema_components(*table_names)
        )
        # create table name conversion lookup table
        name_lookup = {
            folded_name: name for folded_name, name in zip(folded_table_names, table_names)
        }
        # this should never happen: we verify schema for name collisions before loading
        assert len(name_lookup) == len(table_names), (
            f"One or more of tables in {table_names} after applying"
            f" {self.capabilities.casefold_identifier} produced a name collision."
        )
        # if we have more tables to lookup than a threshold, we prefer to filter them in code
        if (
            len(name_lookup) > self.config.info_tables_query_threshold
            or len(",".join(folded_table_names)) > self.capabilities.max_query_length / 2
        ):
            logger.info(
                "Fallback to query all columns from INFORMATION_SCHEMA due to limited query length"
                " or table threshold"
            )
            folded_table_names = []

        query, db_params = self._get_info_schema_columns_query(
            catalog_name, schema_name, folded_table_names
        )
        rows = self.sql_client.execute_sql(query, *db_params)
        prev_table: str = None
        storage_columns: TTableSchemaColumns = None
        for c in rows:
            # if we are selecting all tables this is expected
            if not folded_table_names and c[0] not in name_lookup:
                continue
            # make sure that new table is known
            assert (
                c[0] in name_lookup
            ), f"Table name {c[0]} not in expected tables {name_lookup.keys()}"
            table_name = name_lookup[c[0]]
            if prev_table != table_name:
                # yield what we have
                if storage_columns:
                    yield (prev_table, storage_columns)
                # we have new table
                storage_columns = {}
                prev_table = table_name
                # remove from table_names
                table_names.remove(prev_table)
            # add columns
            col_name = c[1]
            numeric_precision = (
                c[4] if self.capabilities.schema_supports_numeric_precision else None
            )
            numeric_scale = c[5] if self.capabilities.schema_supports_numeric_precision else None

            schema_c: TColumnSchemaBase = {
                "name": col_name,
                "nullable": info_schema_null_to_bool(c[3]),
                **self._from_db_type(c[2], numeric_precision, numeric_scale),
            }
            storage_columns[col_name] = schema_c  # type: ignore
        # yield last table, it must have at least one column or we had no rows
        if storage_columns:
            yield (prev_table, storage_columns)
        # if no columns we assume that table does not exist
        for table_name in table_names:
            yield (table_name, {})

    def get_storage_table(self, table_name: str) -> Tuple[bool, TTableSchemaColumns]:
        """Uses get_storage_tables to get single `table_name` schema.

        Returns (True, ...) if table exists and (False, {}) when not
        """
        storage_table = list(self.get_storage_tables([table_name]))[0]
        return len(storage_table[1]) > 0, storage_table[1]

    @abstractmethod
    def _from_db_type(
        self, db_type: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        pass

    def get_stored_schema(self, schema_name: str = None) -> StorageSchemaInfo:
        name = self.sql_client.make_qualified_table_name(self.schema.version_table_name)
        c_schema_name, c_inserted_at = self._norm_and_escape_columns("schema_name", "inserted_at")
        if not schema_name:
            query = (
                f"SELECT {self.version_table_schema_columns} FROM {name}"
                f" ORDER BY {c_inserted_at} DESC"
            )
            return self._row_to_schema_info(query)
        else:
            query = (
                f"SELECT {self.version_table_schema_columns} FROM {name} WHERE {c_schema_name} = %s"
                f" ORDER BY {c_inserted_at} DESC"
            )
            return self._row_to_schema_info(query, schema_name)

    def get_stored_state(self, pipeline_name: str) -> StateInfo:
        state_table = self.sql_client.make_qualified_table_name(self.schema.state_table_name)
        loads_table = self.sql_client.make_qualified_table_name(self.schema.loads_table_name)
        c_load_id, c_dlt_load_id, c_pipeline_name, c_status = self._norm_and_escape_columns(
            C_DLT_LOADS_TABLE_LOAD_ID, C_DLT_LOAD_ID, "pipeline_name", "status"
        )

        maybe_limit_clause_1, maybe_limit_clause_2 = self.sql_client._limit_clause_sql(1)

        query = (
            f"SELECT {maybe_limit_clause_1} {self.state_table_columns} FROM {state_table} AS s JOIN"
            f" {loads_table} AS l ON l.{c_load_id} = s.{c_dlt_load_id} WHERE {c_pipeline_name} = %s"
            f" AND l.{c_status} = 0 ORDER BY {c_load_id} DESC {maybe_limit_clause_2}"
        )
        with self.sql_client.execute_query(query, pipeline_name) as cur:
            row = cur.fetchone()
        if not row:
            return None
        # NOTE: we request order of columns in SELECT statement which corresponds to StateInfo
        return StateInfo(
            version=row[0],
            engine_version=row[1],
            pipeline_name=row[2],
            state=row[3],
            created_at=pendulum.instance(row[4]),
            _dlt_load_id=row[5],
        )

    def _norm_and_escape_columns(self, *columns: str) -> Iterator[str]:
        return map(
            self.sql_client.escape_column_name, map(self.schema.naming.normalize_path, columns)
        )

    def get_stored_schema_by_hash(self, version_hash: str) -> StorageSchemaInfo:
        table_name = self.sql_client.make_qualified_table_name(self.schema.version_table_name)
        (c_version_hash,) = self._norm_and_escape_columns("version_hash")

        maybe_limit_clause_1, maybe_limit_clause_2 = self.sql_client._limit_clause_sql(1)

        query = (
            f"SELECT {maybe_limit_clause_1} {self.version_table_schema_columns} FROM"
            f" {table_name} WHERE {c_version_hash} = %s {maybe_limit_clause_2}"
        )
        return self._row_to_schema_info(query, version_hash)

    def _get_info_schema_columns_query(
        self, catalog_name: Optional[str], schema_name: str, folded_table_names: List[str]
    ) -> Tuple[str, List[Any]]:
        """Generates SQL to query INFORMATION_SCHEMA.COLUMNS for a set of tables in `folded_table_names`. Input identifiers must be already
        in a form that can be passed to a query via db_params. `catalogue_name` and `folded_tableS_name` is optional and when None, the part of query selecting it
        is skipped.

        Returns: query and list of db_params tuple
        """
        query = f"""
SELECT {",".join(self._get_storage_table_query_columns())}
    FROM {self.sql_client._qualify_info_schema_table_name("COLUMNS")}
WHERE """

        db_params = []
        if catalog_name:
            db_params.append(catalog_name)
            query += "table_catalog = %s AND "
        db_params.append(schema_name)
        select_tables_clause = ""
        # look for particular tables only when requested, otherwise return the full schema
        if folded_table_names:
            db_params = db_params + folded_table_names
            # placeholder for each table
            table_placeholders = ",".join(["%s"] * len(folded_table_names))
            select_tables_clause = f"AND table_name IN ({table_placeholders})"
        query += f"table_schema = %s {select_tables_clause} ORDER BY table_name, ordinal_position"

        return query, db_params

    def _get_storage_table_query_columns(self) -> List[str]:
        """Column names used when querying table from information schema.
        Override for databases that use different namings.
        """
        fields = ["table_name", "column_name", "data_type", "is_nullable"]
        if self.capabilities.schema_supports_numeric_precision:
            fields += ["numeric_precision", "numeric_scale"]
        return fields

    def _execute_schema_update_sql(self, only_tables: Iterable[str]) -> TSchemaTables:
        # Only `only_tables` are included, or all if None.
        sql_scripts, schema_update = self._build_schema_update_sql(
            list(self.get_storage_tables(only_tables or self.schema.tables.keys()))
        )
        # Stay within max query size when doing DDL.
        # Some DB backends use bytes not characters, so decrease the limit by half,
        # assuming most of the characters in DDL encoded into single bytes.
        self.sql_client.execute_many(sql_scripts)
        self._update_schema_in_storage(self.schema)
        return schema_update

    def _build_schema_update_sql(
        self, storage_tables: Iterable[Tuple[str, TTableSchemaColumns]]
    ) -> Tuple[List[str], TSchemaTables]:
        """Generates CREATE/ALTER sql for tables that differ between the destination and in the client's Schema.

        This method compares schema tables to the respective tables in the destination passed in `storage_tables`
        It detects only new tables and new columns. Any other changes like data types, hints, etc. are ignored.

        Args:
            storage_tables (Iterable[Tuple[str, TTableSchemaColumns]]): list of storage tables (tuples (name, column schema))

        Returns:
            Tuple[List[str], TSchemaTables]: Tuple with a list of CREATE/ALTER scripts, and a list of all tables with columns that will be added.
        """
        sql_updates = []
        post_sql_updates = []
        schema_update: TSchemaTables = {}
        for table_name, storage_columns in storage_tables:
            # this will skip incomplete columns
            new_columns = self._create_table_update(table_name, storage_columns)
            generate_alter = len(storage_columns) > 0
            if len(new_columns) > 0:
                # build and add sql to execute
                self._check_table_update_hints(table_name, new_columns, generate_alter)
                sql_statements = self._get_table_update_sql(table_name, new_columns, generate_alter)
                for sql in sql_statements:
                    sql_updates.append(sql)
                # create a schema update for particular table
                partial_table = copy(self.prepare_load_table(table_name))
                # keep only new columns
                partial_table["columns"] = {c["name"]: c for c in new_columns}
                schema_update[table_name] = partial_table
                post_sql_statements = self._get_table_post_update_sql(partial_table)
                for sql in post_sql_statements:
                    post_sql_updates.append(sql)

        # add post sql updates at the end
        sql_updates.extend(post_sql_updates)

        return sql_updates, schema_update

    def _make_add_column_sql(
        self, new_columns: Sequence[TColumnSchema], table: PreparedTableSchema = None
    ) -> List[str]:
        """Make one or more ADD COLUMN sql clauses to be joined in ALTER TABLE statement(s)"""
        return [f"ADD COLUMN {self._get_column_def_sql(c, table)}" for c in new_columns]

    def _make_create_table(self, qualified_name: str, table: PreparedTableSchema) -> str:
        """Begins CREATE TABLE statement"""
        not_exists_clause = " "
        if (
            table["name"] in self.schema.dlt_table_names()
            and self.capabilities.supports_create_table_if_not_exists
        ):
            not_exists_clause = " IF NOT EXISTS "
        return f"CREATE TABLE{not_exists_clause}{qualified_name}"

    @staticmethod
    def _make_alter_table(qualified_name: str) -> str:
        """Begins ALTER TABLE statement"""
        return f"ALTER TABLE {qualified_name}\n"

    def _get_table_update_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> List[str]:
        """Generates a list of SQL statements that updates table `table_name` to include `new_columns`
        columns. `generate_alter` is set to True if table already exists in destination.
        """
        qualified_name = self.sql_client.make_qualified_table_name(table_name)
        table = self.prepare_load_table(table_name)
        sql_result: List[str] = []
        if not generate_alter:
            # build CREATE
            sql = self._make_create_table(qualified_name, table) + " (\n"
            sql += ",\n".join([self._get_column_def_sql(c, table) for c in new_columns])
            sql += self._get_constraints_sql(table_name, new_columns, generate_alter)
            sql += ")"
            sql_result.append(sql)
        else:
            sql_base = self._make_alter_table(qualified_name)
            add_column_statements = self._make_add_column_sql(new_columns, table)
            if self.capabilities.alter_add_multi_column:
                column_sql = ",\n"
                sql_result.append(sql_base + column_sql.join(add_column_statements))
            else:
                # build ALTER as a separate statement for each column (redshift limitation)
                sql_result.extend(
                    [sql_base + col_statement for col_statement in add_column_statements]
                )
            constraints_sql = self._get_constraints_sql(table_name, new_columns, generate_alter)
            if constraints_sql:
                sql_result.append(constraints_sql)
        return sql_result

    def _get_constraints_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> str:
        """Creates or alters additional constraints that are not inlined into CREATE/ALTER TABLE statements"""
        return ""

    def _get_table_post_update_sql(self, partial_table: TPartialTableSchema) -> List[str]:
        """Generates SQL statements executed after all tables are migrated i.e. containing foreign reference.
        `partial_table` contains all table hints and new columns with their hints.
        """
        return []

    def _check_table_update_hints(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> None:
        # scan columns to get hints
        if generate_alter:
            # no hints may be specified on added columns
            for hint in COLUMN_HINTS:
                if any(not has_default_column_prop_value(hint, c.get(hint)) for c in new_columns):
                    hint_columns = [
                        self.sql_client.escape_column_name(c["name"])
                        for c in new_columns
                        if not has_default_column_prop_value(hint, c.get(hint))
                    ]
                    # some hints are not materialized in destination or may be materialized in alter
                    # TODO: add this information to TColumnPropInfo
                    if hint in [
                        "hard_delete",
                        "dedup_sort",
                        "merge_key",
                        "variant",
                        "row_key",
                        "parent_key",
                        "root_key",
                    ]:
                        # you may add those
                        pass
                    elif hint == "nullable":
                        logger.warning(
                            f"Column(s) {hint_columns} with NOT NULL are being added to existing"
                            f" table {table_name}. If there's data in the table the operation"
                            " will fail."
                        )
                    else:
                        logger.warning(
                            f"Column(s) {hint_columns} with hint {hint} are being added to existing"
                            f" table {table_name}. Several hint types may not be added to"
                            " existing tables."
                        )

    def _get_column_def_sql(self, c: TColumnSchema, table: PreparedTableSchema = None) -> str:
        hints_ = self._get_column_hints_sql(c)
        column_name = self.sql_client.escape_column_name(c["name"])
        nullability = self._gen_not_null(c.get("nullable", True))
        column_type = self.type_mapper.to_destination_type(c, table)

        return f"{column_name} {column_type} {hints_} {nullability}"

    def _get_column_hints_sql(self, c: TColumnSchema) -> str:
        return " ".join(
            self.active_hints.get(h, "")
            for h in self.active_hints.keys()
            if c.get(h, False) is True  # use ColumnPropInfos to get default value
        )

    def _gen_not_null(self, nullable: bool) -> str:
        return "NOT NULL" if not nullable else ""

    def _create_table_update(
        self, table_name: str, storage_columns: TTableSchemaColumns
    ) -> Sequence[TColumnSchema]:
        """Compares storage columns with schema table and produce delta columns difference"""
        updates = self.schema.get_new_table_columns(
            table_name,
            storage_columns,
            case_sensitive=self.capabilities.generates_case_sensitive_identifiers(),
        )
        logger.info(f"Found {len(updates)} updates for {table_name} in {self.schema.name}")
        return updates

    def _row_to_schema_info(self, query: str, *args: Any) -> StorageSchemaInfo:
        row: Tuple[Any, ...] = None
        # if there's no dataset/schema return none info
        with contextlib.suppress(DatabaseUndefinedRelation):
            with self.sql_client.execute_query(query, *args) as cur:
                row = cur.fetchone()
        if not row:
            return None

        # get schema as string
        # TODO: Re-use decompress/compress_state() implementation from dlt.pipeline.state_sync
        schema_str: str = row[5]
        try:
            schema_bytes = base64.b64decode(schema_str, validate=True)
            schema_str = zlib.decompress(schema_bytes).decode("utf-8")
        except ValueError:
            # not a base64 string
            pass

        # make utc datetime
        inserted_at = pendulum.instance(row[2])

        return StorageSchemaInfo(row[4], row[3], row[0], row[1], inserted_at, schema_str)

    def _delete_schema_in_storage(self, schema: Schema) -> None:
        """
        Delete all stored versions with the same name as given schema.
        Fails silently if versions table does not exist
        """
        name = self.sql_client.make_qualified_table_name(self.schema.version_table_name)
        (c_schema_name,) = self._norm_and_escape_columns("schema_name")
        self.sql_client.execute_sql(f"DELETE FROM {name} WHERE {c_schema_name} = %s", schema.name)

    def _update_schema_in_storage(self, schema: Schema) -> None:
        # get schema string or zip
        schema_str = json.dumps(schema.to_dict())
        # TODO: not all databases store data as utf-8 but this exception is mostly for redshift
        schema_bytes = schema_str.encode("utf-8")
        if len(schema_bytes) > self.capabilities.max_text_data_type_length:
            # compress and to base64
            schema_str = base64.b64encode(zlib.compress(schema_bytes, level=9)).decode("ascii")
        self._commit_schema_update(schema, schema_str)

    def _commit_schema_update(self, schema: Schema, schema_str: str) -> None:
        now_ts = pendulum.now()
        name = self.sql_client.make_qualified_table_name(self.schema.version_table_name)
        # values =  schema.version_hash, schema.name, schema.version, schema.ENGINE_VERSION, str(now_ts), schema_str
        self.sql_client.execute_sql(
            f"INSERT INTO {name}({self.version_table_schema_columns}) VALUES (%s, %s, %s, %s, %s,"
            " %s)",
            schema.version,
            schema.ENGINE_VERSION,
            now_ts,
            schema.name,
            schema.stored_version_hash,
            schema_str,
        )

    def verify_schema(
        self, only_tables: Iterable[str] = None, new_jobs: Iterable[ParsedLoadJobFileName] = None
    ) -> List[PreparedTableSchema]:
        loaded_tables = super().verify_schema(only_tables, new_jobs)
        if exceptions := verify_schema_merge_disposition(
            self.schema, loaded_tables, self.capabilities, warnings=True
        ):
            filtered = []
            for exception in exceptions:
                if (
                    isinstance(exception, WriteDispositionNotSupported)
                    and self.allow_merge_to_append_fallback
                ):
                    # some destinations allow fallback to append so just warn
                    logger.warning(str(exception))
                else:
                    filtered.append(exception)
                    logger.error(str(exception))
            if filtered:
                raise filtered[0]

        if exceptions := verify_schema_replace_disposition(
            self.schema,
            loaded_tables,
            self.capabilities,
            self.config.replace_strategy,
            warnings=True,
        ):
            for exception in exceptions:
                logger.error(str(exception))
            raise exceptions[0]
        return loaded_tables

    def prepare_load_job_execution(self, job: RunnableLoadJob) -> None:
        self._set_query_tags_for_job(load_id=job._load_id, table=job._load_table)

    def _set_query_tags_for_job(self, load_id: str, table: PreparedTableSchema) -> None:
        """Sets query tags in sql_client for a job in package `load_id`, starting for a particular `table`"""
        from dlt.common.pipeline import current_pipeline

        pipeline = current_pipeline()
        pipeline_name = pipeline.pipeline_name if pipeline else ""
        self.sql_client.set_query_tags(
            {
                "source": self.schema.name,
                "resource": (
                    get_inherited_table_hint(
                        self.schema.tables, table["name"], "resource", allow_none=True
                    )
                    or ""
                ),
                "table": table["name"],
                "load_id": load_id,
                "pipeline_name": pipeline_name,
            }
        )


class SqlJobClientWithStagingDataset(SqlJobClientBase, WithStagingDataset):
    in_staging_dataset_mode: bool = False

    @contextlib.contextmanager
    def with_staging_dataset(self) -> Iterator["SqlJobClientBase"]:
        try:
            with self.sql_client.with_staging_dataset():
                self.in_staging_dataset_mode = True
                yield self
        finally:
            self.in_staging_dataset_mode = False

    def should_load_data_to_staging_dataset(self, table_name: str) -> bool:
        table = self.prepare_load_table(table_name)
        if table["write_disposition"] == "merge":
            return True
        elif table["write_disposition"] == "replace" and (
            table["x-replace-strategy"] in ["insert-from-staging", "staging-optimized"]  # type: ignore[typeddict-item]
        ):
            return True
        return False



from typing import (
    Optional,
    Any,
    Type,
    TYPE_CHECKING,
    Literal,
    List,
    Callable,
    Union,
    Dict,
)
from typing_extensions import TypeAlias

from sqlalchemy.exc import NoReferencedTableError

from dlt.common.typing import TypedDict
from dlt.common.libs.sql_alchemy import Table, Column, Row, sqltypes, Select, TypeEngine
from dlt.common import logger
from dlt.common.schema.typing import TColumnSchema, TTableSchemaColumns, TTableReference

ReflectionLevel = Literal["minimal", "full", "full_with_precision"]


# optionally create generics with any so they can be imported by dlt importer
if TYPE_CHECKING:
    SelectAny: TypeAlias = Select[Any]  # type: ignore[type-arg]
    ColumnAny: TypeAlias = Column[Any]  # type: ignore[type-arg]
    RowAny: TypeAlias = Row[Any]  # type: ignore[type-arg]
    TypeEngineAny = TypeEngine[Any]  # type: ignore[type-arg]
else:
    SelectAny: TypeAlias = Type[Any]
    ColumnAny: TypeAlias = Type[Any]
    RowAny: TypeAlias = Type[Any]
    TypeEngineAny = Type[Any]


TTypeAdapter = Callable[[TypeEngineAny], Optional[Union[TypeEngineAny, Type[TypeEngineAny]]]]


def _is_uuid_type(sql_t: TypeEngineAny) -> bool:
    """Detect UUID-like types across SQLAlchemy 1.4 and 2.0.

    SA 2.0 has a common sqltypes.Uuid base for PG UUID, MSSQL UNIQUEIDENTIFIER, etc.
    SA 1.4 has none of that: PG UUID exposes `as_uuid` but MSSQL UNIQUEIDENTIFIER
    has neither `as_uuid` nor a Uuid base class — only its __visit_name__ identifies it.
    """
    # SA 2.0: common Uuid base type
    if hasattr(sqltypes, "Uuid") and isinstance(sql_t, sqltypes.Uuid):
        return True
    # SA 1.4: PG UUID has as_uuid but no Uuid base
    if hasattr(sql_t, "as_uuid"):
        return True
    # SA 1.4: MSSQL UNIQUEIDENTIFIER has neither as_uuid nor Uuid base
    if getattr(sql_t, "__visit_name__", None) == "UNIQUEIDENTIFIER":
        return True
    return False


class TReflectedHints(TypedDict, total=False):
    columns: TTableSchemaColumns
    references: Optional[List[TTableReference]]
    primary_key: Optional[List[str]]


def default_table_adapter(
    table: Table,
    included_columns: Optional[List[str]],
    excluded_columns: Optional[List[str]] = None,
) -> None:
    """Default table adapter being always called before custom one"""
    if included_columns is not None:
        # Delete columns not included in the load
        for col in list(table._columns):  # type: ignore[attr-defined]
            if col.name not in included_columns:
                table._columns.remove(col)  # type: ignore[attr-defined]

    if excluded_columns is not None:
        for col in list(table._columns):  # type: ignore[attr-defined]
            if col.name in excluded_columns:
                table._columns.remove(col)  # type: ignore[attr-defined]

    for col in table._columns:  # type: ignore[attr-defined]
        sql_t = col.type
        if _is_uuid_type(sql_t) and hasattr(sql_t, "as_uuid"):
            # emit uuids as string by default (PG UUID and SA 2.0 Uuid have as_uuid,
            # MSSQL UNIQUEIDENTIFIER on SA 1.4 does not — pyodbc already returns strings)
            sql_t.as_uuid = False


def sqla_col_to_column_schema(
    sql_col: ColumnAny,
    reflection_level: ReflectionLevel,
    type_adapter_callback: Optional[TTypeAdapter] = None,
    skip_nested_columns_on_minimal: bool = False,
) -> Optional[TColumnSchema]:
    """Infer dlt schema column type from an sqlalchemy type.

    If `add_precision` is set, precision and scale is inferred from that types that support it,
    such as numeric, varchar, int, bigint. Numeric (decimal) types have always precision added.
    """
    col: TColumnSchema = {
        "name": sql_col.name,
    }
    # NOTE: nullability info may not be available for subquery columns
    if hasattr(sql_col, "nullable") and sql_col.nullable is not None:
        col["nullable"] = sql_col.nullable

    if reflection_level == "minimal":
        # normalized into subtables
        if (
            isinstance(sql_col.type, (sqltypes.JSON, sqltypes.ARRAY))
            and skip_nested_columns_on_minimal
        ):
            return None
        return col

    sql_t = sql_col.type

    if type_adapter_callback:
        sql_t = type_adapter_callback(sql_t)
        # Check if sqla type class rather than instance is returned
        if sql_t is not None and isinstance(sql_t, type):
            sql_t = sql_t()

    if sql_t is None:
        # Column ignored by callback
        return col

    add_precision = reflection_level == "full_with_precision"

    if _is_uuid_type(sql_t):
        # we represent UUID as text by default, see default_table_adapter
        col["data_type"] = "text"
    elif isinstance(sql_t, sqltypes.Numeric):
        # check for Numeric type first and integer later, some numeric types (ie. Oracle)
        # derive from both
        # all Numeric types that are returned as floats will assume "double" type
        # and returned as decimals will assume "decimal" type
        if sql_t.asdecimal is False:
            col["data_type"] = "double"
        else:
            col["data_type"] = "decimal"
            if sql_t.precision is not None:
                col["precision"] = sql_t.precision
                # must have a precision for any meaningful scale
                if sql_t.scale is not None:
                    col["scale"] = sql_t.scale
                elif sql_t.decimal_return_scale is not None:
                    col["scale"] = sql_t.decimal_return_scale
    elif isinstance(sql_t, sqltypes.SmallInteger):
        col["data_type"] = "bigint"
        if add_precision:
            col["precision"] = 32
    elif isinstance(sql_t, sqltypes.Integer):
        col["data_type"] = "bigint"
    elif isinstance(sql_t, sqltypes.String):
        col["data_type"] = "text"
        if add_precision and sql_t.length:
            col["precision"] = sql_t.length
    elif isinstance(sql_t, sqltypes._Binary):
        col["data_type"] = "binary"
        if add_precision and sql_t.length:
            col["precision"] = sql_t.length
    elif isinstance(sql_t, sqltypes.DateTime):
        col["data_type"] = "timestamp"
        # special handling for MSSQL
        col["timezone"] = sql_t.timezone or sql_t.__visit_name__ in ("DATETIMEOFFSET")
    elif isinstance(sql_t, sqltypes.Date):
        col["data_type"] = "date"
    elif isinstance(sql_t, sqltypes.Time):
        col["data_type"] = "time"
    elif isinstance(sql_t, sqltypes.JSON):
        col["data_type"] = "json"
    elif isinstance(sql_t, sqltypes.ARRAY):
        col["data_type"] = "json"
    elif isinstance(sql_t, sqltypes.Boolean):
        col["data_type"] = "bool"
    else:
        logger.warning(
            f"A column with name {sql_col.name} contains unknown data type {sql_t} which cannot be"
            " mapped to `dlt` data type. When using sqlalchemy backend such data will be passed to"
            " the normalizer. In case of `pyarrow` and `pandas` backend, data types are detected"
            " from numpy ndarrays. In case of other backends, the behavior is backend-specific."
        )
    return {key: value for key, value in col.items() if value is not None}  # type: ignore[return-value]


def get_primary_key(table: Table) -> Optional[List[str]]:
    """Create primary key or return None if no key defined"""
    primary_key = [c.name for c in table.primary_key]
    return primary_key if len(primary_key) > 0 else None


def table_to_columns(
    table: Table,
    reflection_level: ReflectionLevel = "full",
    type_conversion_fallback: Optional[TTypeAdapter] = None,
    skip_nested_columns_on_minimal: bool = False,
) -> TTableSchemaColumns:
    """Convert an sqlalchemy table to a dlt table schema."""
    return {
        col["name"]: col
        for col in (
            sqla_col_to_column_schema(
                c, reflection_level, type_conversion_fallback, skip_nested_columns_on_minimal
            )
            for c in table.columns
        )
        if col is not None
    }


def get_table_references(table: Table) -> Optional[List[TTableReference]]:
    """Resolve table references from SQLAlchemy foreign key constraints in the table"""
    ref_tables: Dict[str, TTableReference] = {}
    for fk_constraint in table.foreign_key_constraints:
        try:
            referenced_table = fk_constraint.referred_table.name
        except NoReferencedTableError as e:
            logger.warning(
                "Foreign key constraint from table %s could not be resolved to a referenced table,"
                " error message: %s",
                table.name,
                e,
            )
            continue

        if fk_constraint.referred_table.schema != table.schema:
            continue

        elements = fk_constraint.elements
        referenced_columns = [element.column.name for element in elements]
        columns = [col.name for col in fk_constraint.columns]
        if referenced_table in ref_tables:
            # Merge multiple foreign keys to the same table
            existing_ref = ref_tables[referenced_table]
            existing_ref["columns"].extend(columns)  # type: ignore[attr-defined]
            existing_ref["referenced_columns"].extend(referenced_columns)  # type: ignore[attr-defined]
        else:
            ref_tables[referenced_table] = {
                "referenced_table": referenced_table,
                "referenced_columns": referenced_columns,
                "columns": columns,
            }
    return list(ref_tables.values())


def table_to_resource_hints(
    table: Table,
    reflection_level: ReflectionLevel = "full",
    type_conversion_fallback: Optional[TTypeAdapter] = None,
    skip_nested_columns_on_minimal: bool = False,
    resolve_foreign_keys: bool = False,
) -> TReflectedHints:
    result: TReflectedHints = {
        "columns": table_to_columns(
            table,
            reflection_level,
            type_conversion_fallback,
            skip_nested_columns_on_minimal,
        ),
        "primary_key": get_primary_key(table),
    }
    if resolve_foreign_keys:
        result["references"] = get_table_references(table)
    # print("RES HINTS", result)
    return result



import sys
from typing import Any, Sequence, Type, cast, List, Dict, Tuple
import argparse
import rich_argparse
from rich.markdown import Markdown

from dlt.version import __version__
from dlt.common.runners import Venv

from dlt._workspace.cli import SupportsCliCommand, echo as fmt, _debug
from dlt._workspace.cli.exceptions import CliCommandException
from dlt._workspace.cli._telemetry_command import (
    telemetry_change_status_command_wrapper,
)
from dlt._workspace.cli.echo import maybe_no_stdin
from dlt._workspace.cli.utils import display_run_context_info

ACTION_EXECUTED = False
DEFAULT_DOCS_URL = "https://dlthub.com/docs/intro"


def print_help(parser: argparse.ArgumentParser) -> None:
    if not ACTION_EXECUTED:
        parser.print_help()


class TelemetryAction(argparse.Action):
    def __init__(
        self,
        option_strings: Sequence[str],
        dest: Any = argparse.SUPPRESS,
        default: Any = argparse.SUPPRESS,
        help: str = None,  # noqa
    ) -> None:
        super(TelemetryAction, self).__init__(
            option_strings=option_strings, dest=dest, default=default, nargs=0, help=help
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str = None,
    ) -> None:
        global ACTION_EXECUTED

        ACTION_EXECUTED = True
        telemetry_change_status_command_wrapper(option_string == "--enable-telemetry")


class NonInteractiveAction(argparse.Action):
    def __init__(
        self,
        option_strings: Sequence[str],
        dest: Any = argparse.SUPPRESS,
        default: Any = argparse.SUPPRESS,
        help: str = None,  # noqa
    ) -> None:
        super(NonInteractiveAction, self).__init__(
            option_strings=option_strings, dest=dest, default=default, nargs=0, help=help
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str = None,
    ) -> None:
        fmt.ALWAYS_CHOOSE_DEFAULT = True
        fmt.note(
            "Non interactive mode. Default choices are automatically made for confirmations and"
            " prompts."
        )


class DebugAction(argparse.Action):
    def __init__(
        self,
        option_strings: Sequence[str],
        dest: Any = argparse.SUPPRESS,
        default: Any = argparse.SUPPRESS,
        help: str = None,  # noqa
    ) -> None:
        super(DebugAction, self).__init__(
            option_strings=option_strings, dest=dest, default=default, nargs=0, help=help
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str = None,
    ) -> None:
        # will show stack traces (and maybe more debug things)
        _debug.enable_debug()


def _create_parser() -> Tuple[argparse.ArgumentParser, Dict[str, SupportsCliCommand]]:
    parser = argparse.ArgumentParser(
        description=(
            "Creates, adds, inspects and deploys dlt pipelines. Further help is available at"
            " https://dlthub.com/docs/reference/command-line-interface."
        ),
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s {version}".format(version=__version__)
    )
    parser.add_argument(
        "--disable-telemetry",
        action=TelemetryAction,
        help="Disables telemetry before command is executed",
    )
    parser.add_argument(
        "--enable-telemetry",
        action=TelemetryAction,
        help="Enables telemetry before command is executed",
    )
    parser.add_argument(
        "--non-interactive",
        action=NonInteractiveAction,
        help=(
            "Non interactive mode. Default choices are automatically made for confirmations and"
            " prompts."
        ),
    )
    parser.add_argument(
        "--debug",
        action=DebugAction,
        help=(
            "Displays full stack traces on exceptions. Useful for debugging if the output is not"
            " clear enough."
        ),
    )
    parser.add_argument(
        "--no-pwd",
        default=False,
        action="store_true",
        help=(
            "Do not add current working directory to sys.path. By default $pwd is added to "
            "reproduce Python behavior when running scripts."
        ),
    )
    subparsers = parser.add_subparsers(title="Available subcommands", dest="command")

    # load plugins
    from dlt.common.configuration import plugins

    m = plugins.manager()
    commands = cast(List[Type[SupportsCliCommand]], m.hook.plug_cli())

    # install Available subcommands
    installed_commands: Dict[str, SupportsCliCommand] = {}
    for c in commands:
        command = c()
        if command.command in installed_commands.keys():
            continue
        command_parser = subparsers.add_parser(
            command.command,
            help=command.help_string,
            description=command.description if hasattr(command, "description") else None,
        )
        command.configure_parser(command_parser)
        installed_commands[command.command] = command

    # recursively add formatter class
    def add_formatter_class(parser: argparse.ArgumentParser) -> None:
        parser.formatter_class = rich_argparse.RichHelpFormatter

        # NOTE: make markup available for console output
        if parser.description:
            parser.description = Markdown(parser.description, style="argparse.text")  # type: ignore
        for action in parser._actions:
            if isinstance(action, argparse._SubParsersAction):
                for _subcmd, subparser in action.choices.items():
                    add_formatter_class(subparser)

    add_formatter_class(parser)

    return parser, installed_commands


def main() -> int:
    parser, installed_commands = _create_parser()
    args = parser.parse_args()

    if Venv.is_virtual_env() and not Venv.is_venv_activated():
        fmt.warning(
            "You are running dlt installed in the global environment, however you have virtual"
            " environment activated. The dlt command will not see dependencies from virtual"
            " environment. You should uninstall the dlt from global environment and install it in"
            " the current virtual environment instead."
        )

    if cmd := installed_commands.get(args.command):
        try:
            # switch to non-interactive if tty not connected
            with maybe_no_stdin():
                display_run_context_info()
                if not args.no_pwd:
                    if "" not in sys.path:
                        sys.path.insert(0, "")
                cmd.execute(args)
        except Exception as ex:
            docs_url = cmd.docs_url if hasattr(cmd, "docs_url") else DEFAULT_DOCS_URL
            error_code = -1
            raiseable_exception = ex

            # overwrite some values if this is a CliCommandException
            if isinstance(ex, CliCommandException):
                error_code = ex.error_code
                docs_url = ex.docs_url or docs_url
                raiseable_exception = ex.raiseable_exception

            # print exception if available
            if raiseable_exception:
                fmt.secho(str(ex), err=True, fg="red")

            fmt.note("Please refer to our docs at '%s' for further assistance." % docs_url)
            if _debug.is_debug_enabled() and raiseable_exception:
                raise raiseable_exception

            return error_code
    else:
        print_help(parser)
        return -1

    return 0


def _main() -> None:
    """Script entry point"""
    exit(main())


if __name__ == "__main__":
    exit(main())



from functools import wraps
import re
from typing import (
    Any,
)

from lancedb.exceptions import MissingValueError, MissingColumnError

from dlt.common.destination.exceptions import (
    DestinationUndefinedEntity,
    DestinationTransientException,
)
from dlt.common.destination.client import JobClientBase
from dlt.common.typing import TFun

lancedb_not_found_pattern = re.compile(
    r"(?i)(not\s+found|unknown\s+table|missing\s+value|missing\s+column)"
)


def is_lancedb_not_found_error(error_message: str) -> bool:
    """Check if the error message indicates a LanceDB not found error."""
    return bool(lancedb_not_found_pattern.search(error_message))


def lancedb_error(f: TFun) -> TFun:
    @wraps(f)
    def _wrap(self: JobClientBase, *args: Any, **kwargs: Any) -> Any:
        try:
            return f(self, *args, **kwargs)
        except ValueError as e:
            if is_lancedb_not_found_error(str(e)):
                raise DestinationUndefinedEntity(e) from e
            raise
        except (
            MissingValueError,
            MissingColumnError,
        ) as status_ex:
            raise DestinationUndefinedEntity(status_ex) from status_ex
        except Exception as e:
            raise DestinationTransientException(e) from e

    return _wrap  # type: ignore[return-value]



from typing import ClassVar, List, Union, Optional

from dlt.common.typing import TSecretStrValue
from dlt.common.configuration.specs.base_configuration import CredentialsConfiguration, configspec


@configspec
class OAuth2Credentials(CredentialsConfiguration):
    client_id: str = None
    client_secret: TSecretStrValue = None
    refresh_token: Optional[TSecretStrValue] = None
    scopes: Optional[List[str]] = None

    token: Optional[TSecretStrValue] = None
    """Access token"""

    # add refresh_token when generating config samples
    __config_gen_annotations__: ClassVar[List[str]] = ["refresh_token"]

    def auth(self, scopes: Union[str, List[str]] = None, redirect_url: str = None) -> None:
        """Authorizes the client using the available credentials

        Uses the `refresh_token` grant if refresh token is available. Note that `scopes` and `redirect_url` are ignored in this flow.
        Otherwise obtains refresh_token via web flow and authorization code grant.

        Sets `token` and `access_token` fields in the credentials on successful authorization.

        Args:
            scopes (Union[str, List[str]], optional): Additional scopes to add to configured scopes. To be used in web flow. Defaults to None.
            redirect_url (str, optional): Redirect url in case of web flow. Defaults to None.
        """
        raise NotImplementedError()

    def add_scopes(self, scopes: Union[str, List[str]]) -> None:
        if not self.scopes:
            if isinstance(scopes, str):
                self.scopes = [scopes]
            else:
                self.scopes = scopes
        else:
            if isinstance(scopes, str):
                if scopes not in self.scopes:
                    self.scopes += [scopes]
            elif scopes:
                self.scopes = list(set(self.scopes + scopes))



import os
import base64
import dataclasses
from datetime import date, datetime, time  # noqa: I251
from typing import Any, Callable, List, Protocol, IO, Union, Dict
from uuid import UUID
from enum import Enum

try:
    from pydantic import BaseModel as PydanticBaseModel
except ImportError:
    PydanticBaseModel = None  # type: ignore[misc]

from dlt.common import known_env
from dlt.common.exceptions import TypeErrorWithKnownTypes
from dlt.common.pendulum import pendulum
from dlt.common.arithmetics import Decimal
from dlt.common.wei import Wei
from dlt.common.utils import map_nested_values_in_place
from dlt.common.libs.hexbytes import HexBytes

TPuaDecoders = List[Callable[[Any], Any]]

JsonSerializable = Union[str, Dict[str, Any]]
"""
Type representing a JSON-serializable object.
"""

JsonEncoder = Callable[[Any], JsonSerializable]
"""
A callable that takes an object and returns a JSON-serializable representation of it.
Should raise `TypeError` if the object cannot be serialized.
"""


_custom_encoder: Union[None, JsonEncoder] = None
"""
Holds the custom encoder function, if set.
This is used as a last-resort fallback for encoding objects.
"""


def _custom_encode(obj: Any) -> JsonSerializable:
    """Returns a JSON-serializable representation of `obj`"""
    global _custom_encoder
    if isinstance(obj, Decimal):
        # always return decimals as string so they are not deserialized back to float
        return str(obj)
    # this works both for standard datetime and pendulum
    elif isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, date):
        return obj.isoformat()
    elif isinstance(obj, time):
        return obj.isoformat()
    elif isinstance(obj, UUID):
        return str(obj)
    elif isinstance(obj, HexBytes):
        return obj.hex()
    elif isinstance(obj, bytes):
        return base64.b64encode(obj).decode("ascii")
    elif hasattr(obj, "asdict"):
        return obj.asdict()  # type: ignore
    elif hasattr(obj, "_asdict"):
        return obj._asdict()  # type: ignore
    elif PydanticBaseModel and isinstance(obj, PydanticBaseModel):
        return obj.model_dump()
    elif dataclasses.is_dataclass(obj):
        return dataclasses.asdict(obj)  # type: ignore
    elif isinstance(obj, Enum):
        return obj.value  # type: ignore[no-any-return]
    elif _custom_encoder is not None:
        return _custom_encoder(obj)
    raise TypeError(f"`{repr(obj)}` is not JSON serializable")


custom_encode: JsonEncoder = _custom_encode

# use PUA range to encode additional types
PUA_START = int(os.environ.get(known_env.DLT_JSON_TYPED_PUA_START, "0xf026"), 16)

_DECIMAL = chr(PUA_START)
_DATETIME = chr(PUA_START + 1)
_DATE = chr(PUA_START + 2)
_UUIDT = chr(PUA_START + 3)
_HEXBYTES = chr(PUA_START + 4)
_B64BYTES = chr(PUA_START + 5)
_WEI = chr(PUA_START + 6)
_TIME = chr(PUA_START + 7)

PUA_START_UTF8_MAGIC = _DECIMAL.encode("utf-8")[:2]


def _datetime_decoder(obj: str) -> pendulum.DateTime:
    if obj.endswith("Z"):
        # Backwards compatibility for data encoded with previous dlt version
        # fromisoformat does not support Z suffix (until py3.11)
        obj = obj[:-1] + "+00:00"
    # tz=None sets no timezone if if it not specified on string
    dt = pendulum.parse(obj, tz=None)
    if not isinstance(dt, pendulum.DateTime):
        raise TypeErrorWithKnownTypes("obj", dt, ["pendulum.DateTime"])

    return dt


# define decoder for each prefix
DECODERS: TPuaDecoders = [
    Decimal,
    _datetime_decoder,
    pendulum.Date.fromisoformat,
    UUID,
    HexBytes,
    base64.b64decode,
    Wei,
    pendulum.Time.fromisoformat,
]
# Alternate decoders that decode date/time/datetime to stdlib types instead of pendulum
PY_DATETIME_DECODERS = list(DECODERS)
PY_DATETIME_DECODERS[1] = datetime.fromisoformat
PY_DATETIME_DECODERS[2] = date.fromisoformat
PY_DATETIME_DECODERS[7] = time.fromisoformat
# how many decoders?
PUA_CHARACTER_MAX = len(DECODERS)


def _custom_pua_encode(obj: Any) -> JsonSerializable:
    global _custom_encoder
    # wei is subclass of decimal and must be checked first
    if isinstance(obj, Wei):
        return _WEI + str(obj)
    elif isinstance(obj, Decimal):
        return _DECIMAL + str(obj)
    # this works both for standard datetime and pendulum
    elif isinstance(obj, datetime):
        return _DATETIME + obj.isoformat()
    elif isinstance(obj, date):
        return _DATE + obj.isoformat()
    elif isinstance(obj, time):
        return _TIME + obj.isoformat()
    elif isinstance(obj, UUID):
        return _UUIDT + str(obj)
    elif isinstance(obj, HexBytes):
        return _HEXBYTES + obj.hex()
    elif isinstance(obj, bytes):
        return _B64BYTES + base64.b64encode(obj).decode("ascii")
    elif hasattr(obj, "asdict"):
        return obj.asdict()  # type: ignore[no-any-return]
    elif hasattr(obj, "_asdict"):
        return obj._asdict()  # type: ignore[no-any-return]
    elif dataclasses.is_dataclass(obj):
        return dataclasses.asdict(obj)  # type: ignore[arg-type]
    elif PydanticBaseModel and isinstance(obj, PydanticBaseModel):
        return obj.dict(by_alias=True)
    elif isinstance(obj, Enum):
        # Enum value is just int or str
        return obj.value  # type: ignore[no-any-return]
    elif _custom_encoder is not None:
        return _custom_encoder(obj)
    raise TypeError(f"`{repr(obj)}` is not JSON serializable")


custom_pua_encode: JsonEncoder = _custom_pua_encode


def custom_pua_decode(obj: Any, decoders: TPuaDecoders = DECODERS) -> Any:
    if isinstance(obj, str) and len(obj) > 1:
        c = ord(obj[0]) - PUA_START
        # decode only the PUA space defined in DECODERS
        if c >= 0 and c <= PUA_CHARACTER_MAX:
            try:
                return decoders[c](obj[1:])
            except Exception:
                # return strings that cannot be parsed
                # this may be due
                # (1) someone exposing strings with PUA characters to external systems (ie. via API)
                # (2) using custom types ie. DateTime that does not create correct iso strings
                return obj
    return obj


def custom_pua_decode_nested(obj: Any, decoders: TPuaDecoders = DECODERS) -> Any:
    if isinstance(obj, str):
        return custom_pua_decode(obj, decoders)
    elif isinstance(obj, (list, dict)):
        return map_nested_values_in_place(custom_pua_decode, obj, decoders=decoders)
    return obj


def custom_pua_remove(obj: Any) -> Any:
    """Removes the PUA data type marker and leaves the correctly serialized type representation. Unmarked values are returned as-is."""
    if isinstance(obj, str) and len(obj) > 1:
        c = ord(obj[0]) - PUA_START
        # decode only the PUA space defined in DECODERS
        if c >= 0 and c <= PUA_CHARACTER_MAX:
            return obj[1:]
    return obj


def may_have_pua(line: bytes) -> bool:
    """Checks if bytes string contains pua marker"""
    return PUA_START_UTF8_MAGIC in line


class SupportsJson(Protocol):
    """Minimum adapter for different json parser implementations"""

    _impl_name: str
    """Implementation name"""

    def set_custom_encoder(self, encoder: JsonEncoder) -> None: ...

    """
    Set user-defined custom encoder.
    This encoder will be called if none of the built-in encoders can handle the object.
    """

    def dump(
        self, obj: Any, fp: IO[bytes], sort_keys: bool = False, pretty: bool = False
    ) -> None: ...

    def typed_dump(self, obj: Any, fp: IO[bytes], pretty: bool = False) -> None: ...

    def typed_dumps(self, obj: Any, sort_keys: bool = False, pretty: bool = False) -> str: ...

    def typed_loads(self, s: str) -> Any: ...

    def typed_dumpb(self, obj: Any, sort_keys: bool = False, pretty: bool = False) -> bytes: ...

    def typed_loadb(
        self, s: Union[bytes, bytearray, memoryview], decoders: TPuaDecoders = DECODERS
    ) -> Any: ...

    def dumps(self, obj: Any, sort_keys: bool = False, pretty: bool = False) -> str: ...

    def dumpb(self, obj: Any, sort_keys: bool = False, pretty: bool = False) -> bytes: ...

    def load(self, fp: Union[IO[bytes], IO[str]]) -> Any: ...

    def loads(self, s: str) -> Any: ...

    def loadb(self, s: Union[bytes, bytearray, memoryview]) -> Any: ...


def set_custom_encoder_impl(encoder: JsonEncoder) -> None:
    """
    This is passed through to each SupportsJson implementation and mutates the global custom encoder.
    """
    global _custom_encoder
    _custom_encoder = encoder


# pick the right impl
json: SupportsJson = None
if os.environ.get(known_env.DLT_USE_JSON) == "simplejson":
    from dlt.common.json import _simplejson as _json_d

    json = _json_d  # type: ignore[assignment]
else:
    try:
        from dlt.common.json import _orjson as _json_or

        json = _json_or  # type: ignore[assignment]
    except ImportError:
        from dlt.common.json import _simplejson as _json_simple

        json = _json_simple  # type: ignore[assignment]


__all__ = [
    "json",
    "custom_encode",
    "custom_pua_encode",
    "custom_pua_decode",
    "custom_pua_decode_nested",
    "custom_pua_remove",
    "SupportsJson",
    "JsonSerializable",
    "JsonEncoder",
    "may_have_pua",
    "TPuaDecoders",
    "DECODERS",
    "PY_DATETIME_DECODERS",
]



"""Dashboard utility subpackage.

Import submodules directly: utils.pipeline, utils.queries, utils.trace,
utils.schema, utils.formatters, utils.ui, utils.visualization, utils.home,
utils.data_quality.
"""
import dlt._workspace.helpers.dashboard.utils.pipeline as pipeline
import dlt._workspace.helpers.dashboard.utils.queries as queries
import dlt._workspace.helpers.dashboard.utils.trace as trace
import dlt._workspace.helpers.dashboard.utils.schema as schema
import dlt._workspace.helpers.dashboard.utils.formatters as formatters
import dlt._workspace.helpers.dashboard.utils.ui as ui
import dlt._workspace.helpers.dashboard.utils.visualization as visualization
import dlt._workspace.helpers.dashboard.utils.home as home
import dlt._workspace.helpers.dashboard.utils.data_quality as data_quality

__all__ = [
    "pipeline",
    "queries",
    "trace",
    "schema",
    "formatters",
    "ui",
    "visualization",
    "home",
    "data_quality",
]



from dlt.common.normalizers.typing import TJSONNormalizer, TNormalizersConfig
from dlt.common.normalizers.naming import NamingConvention


__all__ = [
    "NamingConvention",
    "TJSONNormalizer",
    "TNormalizersConfig",
]



from typing import Sequence, List
from importlib.metadata import version as pkg_version
from packaging.requirements import Requirement

from dlt.version import DLT_PKG_NAME


class SourceRequirements:
    """Helper class to parse and manipulate entries in source's requirements.txt"""

    dlt_requirement: Requirement
    """Final dlt requirement that may be updated with destination extras"""
    dlt_requirement_base: Requirement
    """Original dlt requirement without extras"""

    def __init__(self, requirements: Sequence[str]) -> None:
        self.parsed_requirements = [Requirement(req) for req in requirements]
        self.dlt_requirement = self._ensure_dlt_requirement()
        # Dlt requirement without extras
        self.dlt_requirement_base = Requirement(str(self.dlt_requirement))
        self.dlt_requirement_base.extras.clear()

    @classmethod
    def from_string(cls, requirements: str) -> "SourceRequirements":
        """Initialize from requirements.txt string, one dependency per line"""
        return cls([line for line in requirements.splitlines() if line])

    def _ensure_dlt_requirement(self) -> Requirement:
        """Find or create dlt requirement"""
        for req in self.parsed_requirements:
            if req.name == DLT_PKG_NAME:
                return req
        req = Requirement(f"{DLT_PKG_NAME}>={self.current_dlt_version()}")
        self.parsed_requirements.append(req)
        return req

    def update_dlt_extras(self, destination_name: str) -> None:
        """Update the dlt requirement to include destination"""
        if not self.dlt_requirement:
            return
        self.dlt_requirement.extras.add(destination_name)

    def current_dlt_version(self) -> str:
        return pkg_version(DLT_PKG_NAME)

    def dlt_version_constraint(self) -> str:
        return str(self.dlt_requirement.specifier)

    def is_installed_dlt_compatible(self) -> bool:
        """Check whether currently installed version is compatible with dlt requirement

        For example, requirements.txt of the source may specify dlt>=0.3.5,<0.4.0
        and we check whether the installed dlt version (e.g. 0.3.6) falls within this range.
        """
        if not self.dlt_requirement:
            return True
        # Lets always accept pre-releases
        spec = self.dlt_requirement.specifier
        spec.prereleases = True
        return self.current_dlt_version() in spec

    def compiled(self) -> List[str]:
        return [str(req) for req in self.parsed_requirements]



import os
import ast
import contextlib
import tomlkit
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    Sequence,
    Literal,
)

import yaml

from dlt.common.configuration.container import Container
from dlt.common.configuration.providers.provider import EXPLICIT_VALUES_PROVIDER_NAME
from dlt.common.json import json
from dlt.common.typing import AnyType, DictStrAny, TAny, is_any_type, get_args, get_origin
from dlt.common.data_types import coerce_value, py_type_to_sc_type
from dlt.common.configuration.providers import EnvironProvider
from dlt.common.configuration.exceptions import ConfigValueCannotBeCoercedException, LookupTrace
from dlt.common.configuration.specs.base_configuration import (
    BaseConfiguration,
    ContainerInjectableContext,
    is_base_configuration_inner_hint,
    configspec,
)


class ResolvedValueTrace(NamedTuple):
    key: str
    value: Any
    default_value: Any
    hint: AnyType
    sections: Sequence[str]
    provider_name: str
    config: BaseConfiguration

    def is_resolved(self) -> bool:
        # explicit values if present are always resolved
        return self.value is not None or self.provider_name == EXPLICIT_VALUES_PROVIDER_NAME


@configspec
class TraceLogContext(ContainerInjectableContext):
    """Stores log of all config resolver traces, per thread so parallel pipelines may log to it"""

    resolved_traces: List[ResolvedValueTrace] = None
    """Traces with resolved values"""
    all_traces: List[ResolvedValueTrace] = None
    """All logged traces"""

    logging_enabled: bool = True
    """Collect logs by default"""

    def clear(self) -> None:
        self.resolved_traces = []
        self.all_traces = []

    def log(self, trace: ResolvedValueTrace) -> None:
        if not self.logging_enabled:
            return
        if trace.is_resolved():
            self.resolved_traces.append(trace)
        self.all_traces.append(trace)

    def __init__(self) -> None:
        self.clear()

    @staticmethod
    def _get_log_as_dict(traces: List[ResolvedValueTrace]) -> Dict[str, ResolvedValueTrace]:
        """Converts logs into layout path - value dict"""
        return {f'{".".join(t.sections)}.{t.key}': t for t in traces}


def deserialize_value(key: str, value: Any, hint: Type[TAny]) -> TAny:
    try:
        if not is_any_type(hint):
            # if deserializing to base configuration, try parse the value
            if is_base_configuration_inner_hint(hint):
                c = hint()
                if isinstance(value, dict):
                    c.update(value)
                else:
                    try:
                        c.parse_native_representation(value)
                    except (ValueError, NotImplementedError):
                        # maybe try again with json parse
                        with contextlib.suppress(ValueError):
                            c_v = json.loads(value)
                            # only lists and dictionaries count
                            if isinstance(c_v, dict):
                                c.update(c_v)
                            else:
                                raise
                return c  # type: ignore

            literal_values: Tuple[Any, ...] = ()
            if get_origin(hint) is Literal:
                # Literal fields are validated against the literal values
                literal_values = get_args(hint)
                hint_origin = type(literal_values[0])
            else:
                hint_origin = hint

            # coerce value
            hint_dt = py_type_to_sc_type(hint_origin)
            value_dt = py_type_to_sc_type(type(value))

            # eval only if value is string and hint is "json"
            if value_dt == "text" and hint_dt == "json":
                if hint_origin is tuple:
                    # use literal eval for tuples
                    value = ast.literal_eval(value)
                else:
                    # use json for sequences and mappings
                    value = json.loads(value)
                # exact types must match
                if not isinstance(value, hint_origin):
                    raise ValueError(value)
            else:
                # for types that are not nested, reuse schema coercion rules
                if value_dt != hint_dt:
                    value = coerce_value(hint_dt, value_dt, value)
                if literal_values and value not in literal_values:
                    raise ConfigValueCannotBeCoercedException(key, value, hint)
        return value  # type: ignore
    except ConfigValueCannotBeCoercedException:
        raise
    except Exception as exc:
        raise ConfigValueCannotBeCoercedException(key, value, hint) from exc


def serialize_value(value: Any) -> str:
    if value is None:
        raise ValueError(value)
    # return literal for tuples
    if isinstance(value, tuple):
        return str(value)
    if isinstance(value, BaseConfiguration):
        try:
            return str(value.to_native_representation())
        except NotImplementedError:
            # no native representation: use dict
            value = dict(value)
    # coerce type to text which will use json for mapping and sequences
    value_dt = py_type_to_sc_type(type(value))
    return coerce_value("text", value_dt, value)  # type: ignore[no-any-return]


def auto_cast(value: str) -> Any:
    """Parse and cast str `value` to bool, int, float and json

    F[f]alse and T[t]rue strings are cast to bool values
    """
    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False
    with contextlib.suppress(ValueError):
        return coerce_value("bigint", "text", value)
    with contextlib.suppress(ValueError):
        return coerce_value("double", "text", value)
    with contextlib.suppress(ValueError):
        c_v = json.loads(value)
        # only lists and dictionaries count
        if isinstance(c_v, (list, dict)):
            return c_v
    return value


def auto_config_fragment(value: str) -> Optional[DictStrAny]:
    """Tries to parse config fragment assuming toml, yaml and json formats

    Only dicts are considered valid fragments.
    None is returned when not a fragment
    """
    try:
        return tomlkit.parse(value).unwrap()
    except ValueError:
        pass
    with contextlib.suppress(Exception):
        c_v = yaml.safe_load(value)
        if isinstance(c_v, dict):
            return c_v
    with contextlib.suppress(ValueError):
        c_v = json.loads(value)
        # only lists and dictionaries count
        if isinstance(c_v, dict):
            return c_v
    return None


def log_traces(
    config: Optional[BaseConfiguration],
    key: str,
    hint: Type[Any],
    value: Any,
    default_value: Any,
    traces: Sequence[LookupTrace],
) -> None:
    # if logger.is_logging() and logger.log_level() == "DEBUG" and config:
    #     logger.debug(f"Field {key} with type {hint} in {type(config).__name__} {'NOT RESOLVED' if value is None else 'RESOLVED'}")
    # print(f"Field {key} with type {hint} in {type(config).__name__} {'NOT RESOLVED' if value is None else 'RESOLVED'}")
    trace_logger = get_resolved_traces()
    for trace in traces:
        trace_logger.log(
            ResolvedValueTrace(
                key,
                trace.value,
                default_value,
                hint,
                trace.sections,
                trace.provider,
                config,
            )
        )


def get_resolved_traces() -> TraceLogContext:
    """Gets trace logging context, per thread, stopped by default"""
    # may create default value
    return Container()[TraceLogContext]


def add_config_to_env(config: BaseConfiguration, sections: Tuple[str, ...] = ()) -> None:
    """Writes values in configuration back into environment using the naming convention of EnvironProvider. Will descend recursively if embedded BaseConfiguration instances are found"""
    if config.__section__:
        sections += (config.__section__,)
    return add_config_dict_to_env(dict(config), sections, overwrite_keys=True)


def add_config_dict_to_env(
    dict_: Mapping[str, Any],
    sections: Tuple[str, ...] = (),
    overwrite_keys: bool = False,
    destructure_dicts: bool = True,
) -> None:
    """Writes values in dict_ back into environment using the naming convention of EnvironProvider. Applies `sections` if specified. Does not overwrite existing keys by default"""
    for k, v in dict_.items():
        if isinstance(v, BaseConfiguration):
            if not v.__section__:
                embedded_sections = sections + (k,)
            else:
                embedded_sections = sections
            add_config_to_env(v, embedded_sections)
        else:
            env_key = EnvironProvider.get_key_name(k, *sections)
            if env_key not in os.environ or overwrite_keys:
                if v is None:
                    os.environ.pop(env_key, None)
                elif isinstance(v, dict) and destructure_dicts:
                    add_config_dict_to_env(
                        v,
                        sections + (k,),
                        overwrite_keys=overwrite_keys,
                        destructure_dicts=destructure_dicts,
                    )
                else:
                    # skip non-serializable fields
                    with contextlib.suppress(TypeError):
                        os.environ[env_key] = serialize_value(v)



import os

from dlt.common.exceptions import DltException
from dlt.common.runtime.exceptions import RunContextNotAvailable


class WorkspaceException(DltException):
    pass


class WorkspaceRunContextNotAvailable(RunContextNotAvailable, WorkspaceException):
    def __init__(self, run_dir: str):
        msg = (
            f"Workspace could not be found in `{run_dir}` (`{os.path.abspath(run_dir)}`). To enable"
            " workspace, create an empty file with the name `.workspace` in `.dlt` settings"
            " folder."
        )
        super().__init__(run_dir, msg)



from typing import ClassVar, Optional

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import BaseConfiguration, known_sections
from dlt.common.normalizers.typing import TNamingConventionReferenceArg
from dlt.common.typing import DictStrAny


@configspec
class SchemaConfiguration(BaseConfiguration):
    # always in section
    __section__: ClassVar[str] = known_sections.SCHEMA

    naming: Optional[TNamingConventionReferenceArg] = None  # Union[str, NamingConvention]
    json_normalizer: Optional[DictStrAny] = None
    allow_identifier_change_on_table_with_data: Optional[bool] = None
    use_break_path_on_normalize: Optional[bool] = None
    """Post 1.4.0 to allow table and column names that contain table separators"""



import logging
from typing import Optional

from dlt.common.exceptions import MissingDependencyException

try:
    import sentry_sdk
    from sentry_sdk.transport import HttpTransport
    from sentry_sdk.integrations.logging import LoggingIntegration
except ModuleNotFoundError:
    raise MissingDependencyException(
        "sentry telemetry",
        ["sentry-sdk"],
        "Please install sentry-sdk if you have `sentry_dsn` set in your RuntimeConfiguration",
    )

from dlt.common.typing import DictStrAny, Any, StrAny
from dlt.common.configuration.specs import RuntimeConfiguration
from dlt.common.runtime.exec_info import dlt_version_info, kube_pod_info, github_info


def init_sentry(config: RuntimeConfiguration) -> None:
    version = dlt_version_info(config.pipeline_name)
    sys_ver = version["dlt_version"]
    release = sys_ver + "_" + version.get("commit_sha", "")
    _SentryHttpTransport.timeout = config.request_timeout
    # TODO: ignore certain loggers ie. dbt loggers
    # https://docs.sentry.io/platforms/python/guides/logging/
    sentry_sdk.init(
        config.sentry_dsn,
        before_send=before_send,  # type: ignore
        traces_sample_rate=1.0,
        # disable tornado, boto3, sql alchemy etc.
        auto_enabling_integrations=False,
        integrations=[_get_sentry_log_level(config)],
        release=release,
        transport=_SentryHttpTransport,
    )
    # add version tags
    for k, v in version.items():
        sentry_sdk.set_tag(k, v)
    # add kubernetes tags
    pod_tags = kube_pod_info()
    for k, v in pod_tags.items():
        sentry_sdk.set_tag(k, v)
    # add github info
    github_tags = github_info()
    for k, v in github_tags.items():
        sentry_sdk.set_tag(k, v)
    if "github_user" in github_tags:
        sentry_sdk.set_user({"username": github_tags["github_user"]})


def disable_sentry() -> None:
    sentry_sdk.Hub.current.bind_client(None)


def before_send(event: DictStrAny, _unused_hint: Optional[StrAny] = None) -> Optional[DictStrAny]:
    """Called by sentry before sending event. Does nothing, patch this function in the module for custom behavior"""
    return event


class _SentryHttpTransport(HttpTransport):
    timeout: float = 0

    def _get_pool_options(self, *a: Any, **kw: Any) -> DictStrAny:
        rv = HttpTransport._get_pool_options(self, *a, **kw)
        rv["timeout"] = self.timeout
        return rv


def _get_sentry_log_level(config: RuntimeConfiguration) -> LoggingIntegration:
    log_level = logging._nameToLevel[config.log_level]
    event_level = logging.WARNING if log_level <= logging.WARNING else log_level
    return LoggingIntegration(
        level=logging.INFO,  # Capture info and above as breadcrumbs
        event_level=event_level,  # Send errors as events
    )



import logging
from datetime import datetime  # noqa: I251
import traceback
from logging import Logger
from typing import Any, List, Type

from dlt.common.json import json
from dlt.common.typing import DictStrAny, StrAny

EMPTY_VALUE = "-"
JSON_SERIALIZER = lambda log: json.dumps(log)
COMPONENT_ID = EMPTY_VALUE
COMPONENT_NAME = EMPTY_VALUE
COMPONENT_INSTANCE_INDEX = 0

# The list contains all the attributes listed in
# http://docs.python.org/library/logging.html#logrecord-attributes
RECORD_ATTR_SKIP_LIST = [
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "args",
    "funcName",
    "id",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msg",
    "msecs",
    "msecs",
    "message",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "thread",
    "threadName",
    "extra",
    # Also exclude legacy 'props'
    "props",
]

RECORD_ATTR_SKIP_LIST.append("stack_info")
EASY_TYPES = (str, bool, dict, float, int, list, type(None))

_default_formatter: Type[logging.Formatter] = None
_epoch = datetime(1970, 1, 1)


def config_root_logger() -> None:
    """
    You must call this if you are using root logger.
    Make all root logger' handlers produce JSON format
    & remove duplicate handlers for request instrumentation logging.
    Please made sure that you call this after you called "logging.basicConfig() or logging.getLogger()
    """
    global _default_formatter
    update_formatter_for_loggers([logging.root], _default_formatter)


def init(custom_formatter: Type[logging.Formatter] = None) -> None:
    """
    This is supposed to be called only one time.

    If **custom_formatter** is passed, it will (in non-web context) use this formatter over the default.
    """

    global _default_formatter

    if custom_formatter:
        if not issubclass(custom_formatter, logging.Formatter):
            raise ValueError(
                "`custom_formatter` is not subclass of logging.Formatter", custom_formatter
            )

    _default_formatter = custom_formatter if custom_formatter else JSONLogFormatter
    logging._defaultFormatter = _default_formatter()  # type: ignore

    # go to all the initialized logger and update it to use JSON formatter
    existing_loggers = list(map(logging.getLogger, logging.Logger.manager.loggerDict))
    update_formatter_for_loggers(existing_loggers, _default_formatter)


class BaseJSONFormatter(logging.Formatter):
    """
    Base class for JSON formatters
    """

    base_object_common: DictStrAny = {}

    def __init__(self, *args: Any, **kw: Any) -> None:
        super(BaseJSONFormatter, self).__init__(*args, **kw)
        if COMPONENT_ID and COMPONENT_ID != EMPTY_VALUE:
            self.base_object_common["component_id"] = COMPONENT_ID
        if COMPONENT_NAME and COMPONENT_NAME != EMPTY_VALUE:
            self.base_object_common["component_name"] = COMPONENT_NAME
        if COMPONENT_INSTANCE_INDEX and COMPONENT_INSTANCE_INDEX != EMPTY_VALUE:
            self.base_object_common["component_instance_idx"] = COMPONENT_INSTANCE_INDEX

    def format(self, record: logging.LogRecord) -> str:  # noqa
        log_object = self._format_log_object(record)
        return JSON_SERIALIZER(log_object)

    def _format_log_object(self, record: logging.LogRecord) -> DictStrAny:
        utcnow = datetime.utcnow()
        base_obj = {
            "written_at": iso_time_format(utcnow),
            "written_ts": epoch_nano_second(utcnow),
        }
        base_obj.update(self.base_object_common)
        # Add extra fields
        base_obj.update(self._get_extra_fields(record))
        return base_obj

    def _get_extra_fields(self, record: logging.LogRecord) -> StrAny:
        fields: DictStrAny = {}

        if record.args:
            fields["msg"] = record.msg

        for key, value in record.__dict__.items():
            if key not in RECORD_ATTR_SKIP_LIST:
                if isinstance(value, EASY_TYPES):
                    fields[key] = value
                else:
                    fields[key] = repr(value)

        # Always add 'props' to the root of the log, assumes props is a dict
        if hasattr(record, "props") and isinstance(record.props, dict):
            fields.update(record.props)

        return fields


def _sanitize_log_msg(record: logging.LogRecord) -> str:
    return record.getMessage().replace("\n", "_").replace("\r", "_").replace("\t", "_")


class JSONLogFormatter(BaseJSONFormatter):
    """
    Formatter for non-web application log
    """

    def get_exc_fields(self, record: logging.LogRecord) -> StrAny:
        if record.exc_info:
            exc_info = self.format_exception(record.exc_info)
        else:
            exc_info = record.exc_text
        return {
            "exc_info": exc_info,
            "filename": record.filename,
        }

    @classmethod
    def format_exception(cls, exc_info: Any) -> str:
        return "".join(traceback.format_exception(*exc_info)) if exc_info else ""

    def _format_log_object(self, record: logging.LogRecord) -> DictStrAny:
        json_log_object = super(JSONLogFormatter, self)._format_log_object(record)
        json_log_object.update(
            {
                "msg": _sanitize_log_msg(record),
                "type": "log",
                "logger": record.name,
                "thread": record.threadName,
                "level": record.levelname,
                "module": record.module,
                "line_no": record.lineno,
            }
        )

        if record.exc_info or record.exc_text:
            json_log_object.update(self.get_exc_fields(record))

        return json_log_object


def update_formatter_for_loggers(
    loggers_iter: List[Logger], formatter: Type[logging.Formatter]
) -> None:
    """
    :param formatter:
    :param loggers_iter:
    """
    for logger in loggers_iter:
        if not isinstance(logger, Logger):
            raise RuntimeError("%s is not a logging.Logger instance", logger)
        for handler in logger.handlers:
            if not isinstance(handler.formatter, formatter):
                handler.formatter = formatter()


def epoch_nano_second(datetime_: datetime) -> int:
    return int((datetime_ - _epoch).total_seconds()) * 1000000000 + datetime_.microsecond * 1000


def iso_time_format(datetime_: datetime) -> str:
    return "%04d-%02d-%02dT%02d:%02d:%02d.%03dZ" % (
        datetime_.year,
        datetime_.month,
        datetime_.day,
        datetime_.hour,
        datetime_.minute,
        datetime_.second,
        int(datetime_.microsecond / 1000),
    )



from typing import Optional
from dlt.common.exceptions import DltException


class DatasetException(DltException):
    pass


class RelationHasQueryException(DatasetException):
    def __init__(self, attempted_change: str) -> None:
        msg = (
            "This readable relation was created with a provided sql query. You cannot change"
            f" `{attempted_change}`. Please change the orignal sql query."
        )
        super().__init__(msg)


class RelationUnknownColumnException(DatasetException):
    def __init__(self, column_name: str) -> None:
        msg = (
            f"The selected column `{column_name}` is not known in the dlt schema for this relation."
        )
        super().__init__(msg)


class LineageFailedException(DltException):
    def __init__(self, msg: Optional[str] = None, *, resource_name: Optional[str] = None):
        super().__init__(msg)
        self.resource_name = resource_name



from __future__ import annotations

"""Easy access to active pipelines, state, sources and schemas"""

from typing import TYPE_CHECKING

from dlt.common.storages.load_package import (
    load_package_state,
    destination_state,
)
from dlt.common.runtime.run_context import active as _run_context

from dlt.extract.decorators import (
    get_resource,
    get_resource_metrics,
    get_source_schema as _get_source_schema,
    get_source as _get_source,
)
from dlt.extract.state import (
    source_state,
    resource_state,
    get_current_pipe_name as _get_current_pipe_name,
    get_current_pipe as _get_current_pipe,
)

if TYPE_CHECKING:
    from dlt._workspace._workspace_context import WorkspaceRunContext
from dlt.pipeline.pipeline import Pipeline as _Pipeline


def pipeline() -> _Pipeline:
    """Currently active pipeline ie. the most recently created or run"""
    from dlt import _pipeline

    return _pipeline()


def workspace() -> WorkspaceRunContext:
    from dlt._workspace._workspace_context import active

    return active()


state = source_state
source_schema = _get_source_schema
source = _get_source
resource = get_resource
pipe = _get_current_pipe
resource_name = _get_current_pipe_name
run_context = _run_context
resource_metrics = get_resource_metrics

__all__ = [
    "load_package_state",
    "destination_state",
    "source_state",
    "resource_state",
    "source_schema",
    "pipe",
    "resource_name",
    "run_context",
    "source",
    "resource",
    "resource_metrics",
]



from typing import Dict, Optional, Type

from dlt.common.configuration.specs import RuntimeConfiguration
from dlt.common.configuration.specs.base_configuration import ContainerInjectableContext
from dlt.common.configuration.specs.pluggable_run_context import (
    PluggableRunContext,
    RunContextBase,
)


def initialize_runtime(logger_name: str, runtime_config: RuntimeConfiguration) -> None:
    from dlt.common import logger
    from dlt.common.runtime.exec_info import dlt_version_info
    from dlt.common.runtime.telemetry import start_telemetry

    from dlt.sources.helpers import requests

    version = dlt_version_info(runtime_config.pipeline_name)

    # initialize or re-initialize logging with new settings
    logger.LOGGER = logger._create_logger(
        logger_name,
        runtime_config.log_level,
        runtime_config.log_format,
        runtime_config.pipeline_name,
        version,
    )

    # Init or update default requests client config
    requests.init(runtime_config)
    # initialize telemetry
    start_telemetry(runtime_config)


def restore_run_context(
    run_context: RunContextBase,
    worker_contexts: Optional[
        Dict[Type[ContainerInjectableContext], ContainerInjectableContext]
    ] = None,
) -> None:
    """Restores run context and worker-affinity contexts in a worker process.

    Called by process pool workers to restore the execution environment from the
    main process. The `run_context` is used to create a new `PluggableRunContext`,
    while `worker_contexts` are contexts marked with `worker_affinity=True` that
    are placed directly into the container.

    Args:
        run_context: The RunContextBase from the main process. Must have
            `runtime_config` attached.
        worker_contexts: Dict mapping context types to instances, collected via
            `Container.get_worker_contexts()`. Each context is placed into the
            worker's container under its type key.
    """
    from dlt.common.configuration.container import Container

    # make sure runtime configuration is attached
    assert run_context.runtime_config is not None

    container = Container()
    container[PluggableRunContext] = PluggableRunContext(run_context)

    if worker_contexts:
        for spec, ctx in worker_contexts.items():
            container[spec] = ctx



from typing import Any

from dlt.common.typing import NoneType
from dlt.common.exceptions import MissingDependencyException


try:
    from dlt.common.libs.pandas import pandas

    PandaFrame = pandas.DataFrame
except MissingDependencyException:
    PandaFrame = NoneType

try:
    from dlt.common.libs.pyarrow import pyarrow

    ArrowTable, ArrowRecords = pyarrow.Table, pyarrow.RecordBatch
except MissingDependencyException:
    ArrowTable, ArrowRecords = NoneType, NoneType


def wrap_additional_type(data: Any) -> Any:
    """Wraps any known additional type so it is accepted by DltResource"""
    # pass through None: if optional deps are not defined, they fallback to None type
    if data is None:
        return data

    if isinstance(data, (PandaFrame, ArrowTable, ArrowRecords)):
        return [data]

    return data



import abc
from typing import Any, ClassVar, List, Sequence, Tuple, Type, TypeVar

from dlt.common.configuration.container import Container
from dlt.common.configuration.exceptions import ConfigFieldMissingException, LookupTrace
from dlt.common.configuration.providers.provider import ConfigProvider
from dlt.common.configuration.specs import BaseConfiguration, is_base_configuration_inner_hint
from dlt.common.configuration.utils import deserialize_value, log_traces, auto_cast
from dlt.common.configuration.specs import PluggableRunContext
from dlt.common.typing import AnyType, ConfigValue, SecretValue, TSecretValue

TConfigAny = TypeVar("TConfigAny", bound=Any)


class _Accessor(abc.ABC):
    def __getitem__(self, field: str) -> Any:
        value, traces = self._get_value(field)
        if value is None:
            raise ConfigFieldMissingException(BaseConfiguration(), {field: traces})  # type: ignore[dict-item]
        if isinstance(value, str):
            return auto_cast(value)
        else:
            return value

    def __setitem__(self, field: str, value: Any) -> None:
        sections = field.split(".")
        key = sections.pop()
        self.writable_provider.set_value(key, value, None, *sections)

    def __contains__(self, field: str) -> bool:
        try:
            self[field]
            return True
        except KeyError:
            return False

    def get(self, field: str, expected_type: Type[TConfigAny] = None) -> TConfigAny:
        value: TConfigAny
        value, _ = self._get_value(field, expected_type)
        if value is None:
            return None
        # cast to required type
        if expected_type:
            return deserialize_value(field, value, expected_type)
        else:
            return value

    @property
    @abc.abstractmethod
    def config_providers(self) -> Sequence[ConfigProvider]:
        pass

    @property
    @abc.abstractmethod
    def default_type(self) -> AnyType:
        pass

    @property
    @abc.abstractmethod
    def writable_provider(self) -> ConfigProvider:
        pass

    def _get_providers_from_context(self) -> Sequence[ConfigProvider]:
        return Container()[PluggableRunContext].providers.providers

    def _get_value(self, field: str, type_hint: Type[Any] = None) -> Tuple[Any, List[LookupTrace]]:
        # get default hint type, in case of dlt.secrets it it TSecretValue
        type_hint = type_hint or self.default_type
        # split field into sections and a key
        sections = field.split(".")
        key = sections.pop()
        value = None
        traces: List[LookupTrace] = []
        for provider in self.config_providers:
            if provider.is_empty:
                continue
            value, effective_field = provider.get_value(key, type_hint, None, *sections)
            trace = LookupTrace(provider.name, sections, effective_field, value)
            traces.append(trace)
            if value is not None:
                # log trace
                if is_base_configuration_inner_hint(type_hint):
                    config: BaseConfiguration = type_hint  # type: ignore
                else:
                    config = None
                log_traces(config, key, type_hint, value, None, [trace])
                break
        return value, traces

    @staticmethod
    def register_provider(provider: ConfigProvider) -> None:
        """Registers `provider` to participate in the configuration resolution. `provider`
        is added after all existing providers and will be used if all others do not resolve.
        """
        Container()[PluggableRunContext].providers.add_provider(provider)


class _ConfigAccessor(_Accessor):
    """Provides direct access to configured values that are not secrets."""

    # TODO should represent the resolution order
    def __repr__(self) -> str:
        s = "dlt.config("
        s += "\n  config_providers=["
        for p in self.config_providers:
            s += "\n    " + p.__repr__()
        s += "\n  ]"
        s += "\n)"
        return s

    @property
    def config_providers(self) -> Sequence[ConfigProvider]:
        """Return a list of config providers, in lookup order"""
        return [p for p in self._get_providers_from_context()]

    @property
    def default_type(self) -> AnyType:
        return AnyType

    @property
    def writable_provider(self) -> ConfigProvider:
        """find first writable provider that does not support secrets - should be config.toml"""
        return next(
            p
            for p in self._get_providers_from_context()
            if p.is_writable and not p.supports_secrets
        )

    value: ClassVar[Any] = ConfigValue
    "A placeholder that tells dlt to replace it with actual config value during the call to a source or resource decorated function."


class _SecretsAccessor(_Accessor):
    """Provides direct access to secrets."""

    # TODO should represent the resolution order
    def __repr__(self) -> str:
        s = "dlt.secrets("
        s += "\n  config_providers=["
        for p in self.config_providers:
            s += "\n    " + p.__repr__()
        s += "\n  ]"
        s += "\n)"
        return s

    @property
    def config_providers(self) -> Sequence[ConfigProvider]:
        """Return a list of config providers that can hold secrets, in lookup order"""
        return [p for p in self._get_providers_from_context() if p.supports_secrets]

    @property
    def default_type(self) -> AnyType:
        return TSecretValue

    @property
    def writable_provider(self) -> ConfigProvider:
        """find first writable provider that supports secrets - should be secrets.toml"""
        return next(
            p for p in self._get_providers_from_context() if p.is_writable and p.supports_secrets
        )

    value: ClassVar[Any] = SecretValue
    "A placeholder that tells dlt to replace it with actual secret during the call to a source or resource decorated function."


config = _ConfigAccessor()
"""Dictionary-like access to all config values to dlt"""

secrets = _SecretsAccessor()
"""Dictionary-like access to all secrets known known to dlt"""



from typing import Callable, ClassVar, List, Optional, Tuple
from dlt.common.configuration.specs import known_sections

from dlt.common.configuration.specs.base_configuration import ContainerInjectableContext, configspec


@configspec
class ConfigSectionContext(ContainerInjectableContext):
    TMergeFunc = Callable[["ConfigSectionContext", "ConfigSectionContext"], None]

    worker_affinity: ClassVar[bool] = True

    pipeline_name: Optional[str] = None
    sections: Tuple[str, ...] = ()
    merge_style: TMergeFunc = None
    source_state_key: str = None

    def merge(self, existing: "ConfigSectionContext") -> None:
        """Merges existing context into incoming using a merge style function"""
        merge_style_f = self.merge_style or self.prefer_incoming
        merge_style_f(self, existing)

    def source_name(self) -> str:
        """Gets name of a source from `sections`"""
        if self.sections and len(self.sections) == 3 and self.sections[0] == known_sections.SOURCES:
            return self.sections[-1]
        raise ValueError(self.sections)

    def source_section(self) -> str:
        """Gets section of a source from `sections`"""
        if self.sections and len(self.sections) == 3 and self.sections[0] == known_sections.SOURCES:
            return self.sections[1]
        raise ValueError(self.sections)

    @staticmethod
    def prefer_incoming(incoming: "ConfigSectionContext", existing: "ConfigSectionContext") -> None:
        incoming.pipeline_name = incoming.pipeline_name or existing.pipeline_name
        incoming.sections = incoming.sections or existing.sections
        incoming.source_state_key = incoming.source_state_key or existing.source_state_key

    @staticmethod
    def prefer_existing(incoming: "ConfigSectionContext", existing: "ConfigSectionContext") -> None:
        """Prefer existing section context when merging this context before injecting"""
        incoming.pipeline_name = existing.pipeline_name or incoming.pipeline_name
        incoming.sections = existing.sections or incoming.sections
        incoming.source_state_key = existing.source_state_key or incoming.source_state_key

    @staticmethod
    def resource_merge_style(
        incoming: "ConfigSectionContext", existing: "ConfigSectionContext"
    ) -> None:
        """If top level section is same and there are 3 sections it replaces second element (source module) from existing and keeps the 3rd element (name)"""
        incoming.pipeline_name = incoming.pipeline_name or existing.pipeline_name
        if (
            len(incoming.sections) == 3 == len(existing.sections)
            and incoming.sections[0] == existing.sections[0]
        ):
            # existing does not have middle section then keep incoming
            incoming.sections = (
                incoming.sections[0],
                existing.sections[1] or incoming.sections[1],
                incoming.sections[2],
            )
            incoming.source_state_key = existing.source_state_key or incoming.source_state_key
        else:
            incoming.sections = incoming.sections or existing.sections
            incoming.source_state_key = incoming.source_state_key or existing.source_state_key

    def __str__(self) -> str:
        return (
            super().__str__()
            + f": {self.pipeline_name} {self.sections}@{self.merge_style} state['{self.source_state_key}']"
        )



from contextlib import suppress

from dlt.destinations.impl.mssql.sql_client import PyOdbcMsSqlClient
from dlt.destinations.exceptions import DatabaseUndefinedRelation


class SynapseSqlClient(PyOdbcMsSqlClient):
    def drop_tables(self, *tables: str) -> None:
        if not tables:
            return
        # Synapse does not support DROP TABLE IF EXISTS.
        # Workaround: use DROP TABLE and suppress non-existence errors.
        statements = [f"DROP TABLE {self.make_qualified_table_name(table)}" for table in tables]
        for statement in statements:
            with suppress(DatabaseUndefinedRelation):
                self.execute_sql(statement)



from datetime import datetime, timedelta, date  # noqa: I251
from typing import Any, Union

from dlt.common import logger, pendulum
from dlt.common.time import (
    detect_datetime_format,
    ensure_pendulum_date,
    ensure_pendulum_datetime_non_utc,
    datetime_obj_to_str,
)

from . import TCursorValue, LastValueFunc


def _apply_lag_to_value(
    lag: float, value: Any, last_value_func: LastValueFunc[TCursorValue]
) -> Any:
    """Applies lag to a value, in case of `str` types it attempts to return a string
    with the lag applied preserving original format of a datetime/date
    """
    # Determine if the input is originally a string and capture its format
    value_format: str = None
    if isinstance(value, str):
        value_format = detect_datetime_format(value)
        is_str_date = value_format in ("%Y%m%d", "%Y-%m-%d")
        value = (
            ensure_pendulum_date(value) if is_str_date else ensure_pendulum_datetime_non_utc(value)
        )

    # we must have pendulum instance.
    if isinstance(value, date):
        # we didn't convert to pendulum yet
        if not value_format:
            value = (
                ensure_pendulum_datetime_non_utc(value)
                if isinstance(value, datetime)
                else ensure_pendulum_date(value)
            )
        value = _apply_lag_to_datetime(lag, value, last_value_func)
        # go back to string or pass exact type
        value = datetime_obj_to_str(value, value_format) if value_format else value

    elif isinstance(value, (int, float)):
        value = _apply_lag_to_number(lag, value, last_value_func)

    else:
        raise ValueError(
            value,
            f"Lag is not supported for cursor type: {type(value)} with last_value_func:"
            f" {last_value_func}. Strings must parse to DateTime or Date.",
        )

    return value


def _apply_lag_to_datetime(
    lag: float,
    value: pendulum.DateTime,
    last_value_func: LastValueFunc[TCursorValue],
) -> pendulum.DateTime:
    if last_value_func is max:
        lag = -lag

    if isinstance(value, pendulum.DateTime):
        return value.add(seconds=lag)

    return value.add(days=lag)


def _apply_lag_to_number(
    lag: float, value: Union[int, float], last_value_func: LastValueFunc[TCursorValue]
) -> Union[int, float]:
    adjusted_value = value - lag if last_value_func is max else value + lag
    return int(adjusted_value) if isinstance(value, int) else adjusted_value


def apply_lag(
    lag: float,
    initial_value: TCursorValue,
    last_value: TCursorValue,
    last_value_func: LastValueFunc[TCursorValue],
) -> TCursorValue:
    """Applies lag to `last_value` but prevents it to cross `initial_value`: observing order of last_value_func"""
    # Skip lag adjustment to avoid out-of-bounds issues
    lagged_last_value = _apply_lag_to_value(lag, last_value, last_value_func)
    if (
        initial_value is not None
        and last_value_func((initial_value, lagged_last_value)) == initial_value
    ):
        # do not cross initial_value
        return initial_value
    return lagged_last_value  # type: ignore[no-any-return]



from typing import Any, Optional, Literal, Dict, List, Union, cast

from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.typing import TColumnNames
from dlt.destinations.utils import get_resource_for_adapter
from dlt.extract import DltResource
from dlt.extract.items import TTableHintTemplate
from dlt.destinations.impl.databricks.typing import TDatabricksTableSchemaColumns


CLUSTER_HINT: Literal["x-databricks-cluster"] = "x-databricks-cluster"
TABLE_COMMENT_HINT: Literal["x-databricks-table-comment"] = "x-databricks-table-comment"
TABLE_TAGS_HINT: Literal["x-databricks-table-tags"] = "x-databricks-table-tags"
TABLE_PROPERTIES_HINT: Literal["x-databricks-table-properties"] = "x-databricks-table-properties"
COLUMN_COMMENT_HINT: Literal["x-databricks-column-comment"] = "x-databricks-column-comment"
COLUMN_TAGS_HINT: Literal["x-databricks-column-tags"] = "x-databricks-column-tags"


def databricks_adapter(
    data: Any,
    cluster: Union[TColumnNames, Literal["AUTO"]] = None,
    partition: TColumnNames = None,
    table_format: Literal["DELTA", "ICEBERG"] = "DELTA",
    table_comment: Optional[str] = None,
    table_tags: Optional[List[Union[str, Dict[str, str]]]] = None,
    table_properties: Optional[Dict[str, Union[str, int, bool, float]]] = None,
    column_hints: Optional[TDatabricksTableSchemaColumns] = None,
) -> DltResource:
    """
    Prepares data for loading into Databricks.

    This function takes data, which can be raw or already wrapped in a DltResource object,
    and prepares it for Databricks by optionally specifying clustering, partitioning, and table description.

    Args:
        data (Any): The data to be transformed.
            This can be raw data or an instance of DltResource.
            If raw data is provided, the function will wrap it into a `DltResource` object.
        cluster (Union[TColumnNames, Literal["AUTO"]], optional): A column name, list of column names, or "AUTO" to cluster the Databricks table by.
            Use "AUTO" to let Databricks automatically determine the best clustering.
        partition (TColumnNames, optional): A column name or list of column names to partition the Databricks table by.
            Partitioning divides the table into separate files based on the partition column values.
        table_format (Literal["DELTA", "ICEBERG"], optional): The table format to use. Defaults to "DELTA".
            Use "ICEBERG" to create Apache Iceberg tables for better schema evolution and time travel capabilities.
        table_comment (str, optional): A description for the Databricks table.
        table_tags (List[Union[str, Dict[str, str]]], optional): A list of tags for the Databricks table.
            Can contain a mix of strings and key-value pairs as dictionaries.
            Example: ["production", {"environment": "prod"}, "employees"]
        table_properties (Dict[str, Union[str, int, bool, float]], optional): A dictionary of table properties
            to be added to the Databricks table using TBLPROPERTIES. These are key-value pairs for metadata
            and Delta Lake optimization settings. Example: {"delta.appendOnly": True, "delta.logRetentionDuration": "30 days"}
        column_hints (TTableSchemaColumns, optional): A dictionary of column hints.
            Each key is a column name, and the value is a dictionary of hints.
            The supported hints are:
            - `column_comment` - adds a comment to the column. Supports basic markdown format [basic-syntax](https://www.markdownguide.org/cheat-sheet/#basic-syntax).
            - `column_tags` - adds tags to the column. Supports a list of strings and/or key-value pairs.

    Returns:
        A `DltResource` object that is ready to be loaded into Databricks.

    Raises:
        ValueError: If any hint is invalid or none are specified.

    Examples:
        >>> data = [{"name": "Marcel", "description": "Raccoon Engineer", "date_hired": 1700784000}]
        >>> databricks_adapter(data, cluster="date_hired", table_comment="Employee Data",
        ...     table_tags=["production", {"environment": "prod"}, "employees"])
        >>> # Use AUTO clustering
        >>> databricks_adapter(data, cluster="AUTO", table_comment="Auto-clustered table")
        >>> # Use partitioning
        >>> databricks_adapter(data, partition=["year", "month"], cluster="customer_id")
        >>> # Create Iceberg table
        >>> databricks_adapter(data, table_format="ICEBERG", cluster="customer_id")
    """
    resource = get_resource_for_adapter(data)

    additional_table_hints: Dict[str, TTableHintTemplate[Any]] = {}
    additional_column_hints: TDatabricksTableSchemaColumns = {}

    # Handle table format
    if table_format not in ["DELTA", "ICEBERG"]:
        raise ValueError("`table_format` must be either 'DELTA' or 'ICEBERG'.")

    # Store table format at table level (lowercase to match destination format)
    additional_table_hints["table_format"] = table_format.lower()

    if cluster:
        if cluster == "AUTO":
            # Handle AUTO clustering at table level
            additional_table_hints[CLUSTER_HINT] = "AUTO"
        else:
            # Handle specific column clustering
            if isinstance(cluster, str):
                cluster = [cluster]
            if not isinstance(cluster, list):
                raise ValueError(
                    "`cluster` must be a list of column names, a single column name as a string, or"
                    " 'AUTO'."
                )
            for column_name in cluster:
                additional_column_hints[column_name] = {"name": column_name, CLUSTER_HINT: True}  # type: ignore[typeddict-unknown-key]

    if partition:
        if isinstance(partition, str):
            partition = [partition]
        if not isinstance(partition, list):
            raise ValueError(
                "`partition` must be a list of column names or a single column name as a string."
            )
        for column_name in partition:
            if column_name not in additional_column_hints:
                additional_column_hints[column_name] = {"name": column_name}
            additional_column_hints[column_name]["partition"] = True

    if column_hints:
        for column_name, hints in column_hints.items():
            if column_name not in additional_column_hints:
                additional_column_hints[column_name] = {"name": column_name}

            # cast to a generic dict to access keys not in TColumnSchema
            hints_dict = cast(Dict[str, Any], hints)
            if "column_comment" in hints_dict:
                additional_column_hints[column_name][COLUMN_COMMENT_HINT] = hints_dict[  # type: ignore[typeddict-unknown-key]
                    "column_comment"
                ]
            if "column_tags" in hints_dict:
                column_tags = hints_dict["column_tags"]
                if not isinstance(column_tags, list):
                    raise ValueError(
                        "`column_tags` must be a list of strings and/or key-value pairs."
                    )
                for tag in column_tags:
                    if isinstance(tag, str):
                        continue
                    if isinstance(tag, dict) and len(tag) == 1:
                        # Ensure the dictionary has exactly one key-value pair
                        continue
                    raise ValueError(
                        "Each tag must be either a string or a dictionary with a single key-value"
                        " pair."
                    )
                additional_column_hints[column_name][COLUMN_TAGS_HINT] = column_tags  # type: ignore[typeddict-unknown-key]

    if table_comment:
        if not isinstance(table_comment, str):
            raise ValueError(
                "`table_comment` must be string representing Databricks table description."
            )
        additional_table_hints[TABLE_COMMENT_HINT] = table_comment

    if table_tags:
        if not isinstance(table_tags, list):
            raise ValueError("`table_tags` must be a list of strings and/or key-value pairs.")
        for tag in table_tags:
            if isinstance(tag, str):
                continue
            elif isinstance(tag, dict) and len(tag) == 1:
                # Ensure the dictionary has exactly one key-value pair
                continue
            else:
                raise ValueError(
                    "Each tag must be either a string or a dictionary with a single key-value pair."
                )

        additional_table_hints[TABLE_TAGS_HINT] = table_tags

    if table_properties:
        if not isinstance(table_properties, dict):
            raise ValueError("`table_properties` must be a dictionary of key-value pairs.")

        # Reserved keys that should not be used in TBLPROPERTIES
        reserved_keys = {"external", "location", "owner", "provider"}

        for key, value in table_properties.items():
            if not isinstance(key, str):
                raise ValueError("Table property keys must be strings.")

            # Check for reserved keys
            if key.lower() in reserved_keys:
                raise ValueError(
                    f"Table property key '{key}' is reserved and cannot be used. "
                    f"Reserved keys are: {', '.join(reserved_keys)}"
                )

            # Check for keys starting with 'option.'
            if key.startswith("option."):
                raise ValueError(
                    f"Table property key '{key}' starts with 'option.' which is reserved."
                )

            # Validate value types
            if not isinstance(value, (str, int, bool, float)):
                raise ValueError(
                    f"Table property value for key '{key}' must be a string, integer, boolean, or"
                    f" float. Got {type(value).__name__}."
                )

        additional_table_hints[TABLE_PROPERTIES_HINT] = table_properties

    resource.apply_hints(
        columns=cast(TTableSchemaColumns, additional_column_hints),
        additional_table_hints=additional_table_hints,
    )

    return resource
